import base64
import binascii
import urllib.parse
from typing import Optional, Union

from .._new_field_registry import (
    OPNewItemField,
    op_register_new_item_field_type
)
from .item_section import OPSection


class OPNewTOTPUriException(Exception):
    pass


@op_register_new_item_field_type
class OPNewStringField(OPNewItemField):
    """
    A class for creating a new item field of type 'STRING'
    """
    FIELD_TYPE = "STRING"


@op_register_new_item_field_type
class OPNewConcealedField(OPNewItemField):
    """
    A class for creating a new item field of type 'CONCEALED'
    """
    FIELD_TYPE = "CONCEALED"


class OPNewUsernameField(OPNewStringField):
    """
    A class for creating a new specialized item field of type 'STRING' and purpose 'USERNAME'
    """
    FIELD_PURPOSE = "USERNAME"


class OPNewPasswordField(OPNewConcealedField):
    """
    A class for creating a new specialized item field of type 'CONCEALED' and purpose 'PASSWORD'
    """
    FIELD_PURPOSE = "PASSWORD"


class OPNewTOTPUri:
    """
    A class for constructing a properly formed TOTP URI from a secret, account name, and issuer
    """
    # otpauth://totp/<website>:<user>?secret=<secret>&issuer=<issuer>'
    # https://rootprojects.org/authenticator/

    def __init__(self,
                 secret: str,
                 account_name: Optional[str] = None,
                 issuer: Optional[str] = None):
        """
        Create a new OPNewTOTPUri object for use with a OPNewTOTPField

        Parameters
        ----------
        secret: str
            The base32-encoded secret seed for the TOTP generator.
            The base32 padding should be stripped. This value will be
            verified via base32 decoding
        account_name: str, optional
            The name of the account this TOTP is for. Defaults to "secret"
        issuer: str, optional
            The website or issuer for this TOTP object

        Raises
        ------
        OPNewTOTPUriException
            If the secret fails base32 decode verification
        """
        self._secret = secret
        self._issuer = issuer
        self._account = account_name or "secret"
        self._verify_secret()

    def _verify_secret(self):
        # Re-pad the secret and attempt to base32 decode it
        # raise an exception if decoding fails
        secret = self._secret
        missing_padding = len(secret) % 8
        if missing_padding != 0:
            secret += "=" * (8 - missing_padding)
        try:
            base64.b32decode(secret, casefold=True)
        except binascii.Error as e:
            raise OPNewTOTPUriException(
                f"Invalid secret string: base32 decoding {e}")

    def __str__(self):
        issuer = None
        if self._issuer:
            issuer = urllib.parse.quote(self._issuer)
        account = urllib.parse.quote(self._account)
        if issuer:
            label = f"{issuer}:{account}"
        else:
            label = account

        params_dict = {"secret": self._secret}

        if self._issuer:
            params_dict["issuer"] = self._issuer

        params = urllib.parse.urlencode(
            params_dict, quote_via=urllib.parse.quote)
        url_str = f"otpauth://totp/{label}?{params}"
        return url_str


@op_register_new_item_field_type
class OPNewTOTPField(OPNewStringField):
    """
    A class for creating a new item field of type 'OTP'
    """
    FIELD_TYPE = "OTP"

    def __init__(self,
                 field_label: str,
                 totp_value: Union[str, OPNewTOTPUri],
                 field_id: str = None,
                 section: Optional[OPSection] = None):
        """
        Create a new TOTP field object


        Parameters
        ----------
        field_label: str
            The user-visible name of the field
        totp_value: Union[str, OPNewTOTPUri]
            The TOTP URI value for this field. May be a string or a OPNewTOTPUri object
        field_id: str, optional
            The unique identifier for this field. If none is provided, a random one will be generated
        section: OPSection, optional
            The section this field should be associated with. Not all fields are in sections
        """
        if isinstance(totp_value, OPNewTOTPUri):
            totp_value = str(totp_value)
        super().__init__(field_label, totp_value, field_id, section)


class OPNewNetworkPortField(OPNewStringField):
    """
    A class for creating a new item string field for use with numeric port values.
    """

    def __init__(self, field_label: str, value: Union[str, int], field_id: str = None, section: Optional[OPSection] = None):
        """
        Create an OPNewNetworkPortField for use with new item template objects

        Parameters
        ----------
        field_label : str
            The user-visible name of the field
        value : Union[str, int]
            port value for the new database item, by default None
            Discussion:
                The port value may be either an integer or a string representation of an integer.

                If an integer is provided, it will be converted to a string via 'str(value)'

                If a string is provided, it will be validated by converting it to an integer via 'int(value, 0)'.
                If this fails, ValueError is raised.
                The string may be any valid base representation. E.g., 1234 or 0x1234. This representation will
                preserved.

                No attempt is made to ensure the value is in a reasonable range for what a port should be
        field_id : str, optional
            The unique identifier for this field. If none is provided, a random one will be generated. By default None
        section : OPSection, optional
            The section this field should be associated with. Not all fields are in sections, by default None

        Raises
        ------
        ValueError
            _description_
        """
        if isinstance(value, str):
            # convert to int for sanity checking
            # accept any base. int() will raise ValueError if it can't
            # be converted
            _ = int(value, 0)
        elif isinstance(value, int):
            # only convert 'value' to a string
            # if it wasn't already, in order to preserve alternate base representations
            # e.g., 0x4d2
            value = str(value)
        else:
            # if we were passed any other object type this is an error
            raise ValueError(
                f"Invalid value for port: \"{value}\". Must be int or str")

        # NOTE: we're not going to enforce valid, non-negative port ranges
        # user can enter whatever number they like

        super().__init__(field_label, value, field_id, section)



"""
Tests for custom abstract base class support
"""
from typing import Dict, Union

import pytest

from pyonepassword.op_items.item_types._item_base import OPAbstractItem
from pyonepassword.op_items.item_types._item_descriptor_base import (
    OPAbstractItemDescriptor
)

from .fixtures.valid_data import ValidData

VALID_INPUT_ITEM = "example-login-with-fields"


def test_abc_meta_01(valid_data: ValidData):
    """
    Test:
      - instantiating an Abstract Base Class with a method requiring implementation

    Verify:
        - TypeError is raised
    """
    item_dict = valid_data.data_for_name(VALID_INPUT_ITEM)

    with pytest.raises(TypeError):
        OPAbstractItemDescriptor(item_dict)


def test_abc_meta_02(valid_data: ValidData):
    """
    Test:
      - instantiating a class extending an ABC with that does not implement all enforced methods

    Verify:
        - Subclass is still abstract and TypeError is raised
    """
    item_dict = valid_data.data_for_name(VALID_INPUT_ITEM)

    with pytest.raises(TypeError):
        OPAbstractItem(item_dict)


def test_abc_meta_03(valid_data: ValidData):
    """
    Test:
      - instantiating a class extending an ABC with that does not implement all enforced methods

    Verify:
        - Subclass is still abstract and TypeError is raised
    """
    class OPConcreteItem(OPAbstractItem):
        """
        A subclass of an ABC that properly implements the enforced __init__() method
        """

        def __init__(self, item_dict_or_json: Union[Dict, str]):
            super().__init__(item_dict_or_json)

    item_dict = valid_data.data_for_name(VALID_INPUT_ITEM)

    # No error/exception should be raised, since enforced methods have been implemented
    OPConcreteItem(item_dict)



# __future__.annotations, and typing.TYPE_CHECKING
# enable anything imported for type hinting to disappear at run time
from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

# make imports for type-hinting disappear at run-time to avoid
# circular imports.
# this also reduced exercising tested code simply by importing
if TYPE_CHECKING:
    # non-runtime imports here
    from pyonepassword import OP
    from .fixtures.expected_item_list import ExpectedItemListData, ExpectedItemList

from pyonepassword.api.exceptions import OPItemDeleteMultipleException
from pyonepassword.api.object_types import OPItemList

# ensure HOME env variable is set, and there's a valid op config present
pytestmark = pytest.mark.usefixtures("valid_op_cli_config_homedir")


def _validate_item_list(item_list, expected_item_list):
    assert isinstance(item_list, OPItemList)
    assert len(item_list) == len(expected_item_list)


def test_item_delete_multiple_01(signed_in_op: OP, expected_item_list_data: ExpectedItemListData):
    expected_items: ExpectedItemList = expected_item_list_data.data_for_name(
        "item-list-test-vault-3")
    vault_name = "Test Data 3"
    deleted_items: OPItemList
    deleted_items = signed_in_op.item_delete_multiple(vault_name)

    _validate_item_list(deleted_items, expected_items)

    for expected, actual in zip(expected_items, deleted_items):
        assert expected.unique_id == actual.unique_id
        assert expected.title == actual.title


@pytest.mark.usefixtures("setup_stateful_item_delete_multiple")
def test_item_delete_multiple_02(signed_in_op: OP):
    """
    A test to delete all items of a certain tag and verify none of those items
    remain afterwards

    retrieve list of all items in the test vault tagged 'tag_1'
    delete all items tagged 'tag_1'

    verify:
        - before deleting, there is one or more items tagged 'tag_1'
        - after deleting, there are no items tagged 'tag_1'
    """
    tags = ["tag_1"]
    vault_name = "Test Data 3"

    orig_item_list = signed_in_op.item_list(vault=vault_name, tags=tags)
    assert len(orig_item_list)

    signed_in_op.item_delete_multiple(vault_name, tags=tags)
    new_item_list = signed_in_op.item_list(vault=vault_name, tags=tags)

    assert len(new_item_list) == 0


@pytest.mark.usefixtures("setup_stateful_item_delete_multiple")
def test_item_delete_multiple_03(signed_in_op: OP):
    """
    A test to delete all items of a certain tag, and to verify not all items
    were deleted

    retrieve list of all items in the test vault
    delete all items tagged 'tag_1'
    retrieve list of all remaining items in the test

    verify:
        - before deleting, there is at least one item in the test vault
        - after deleting, there is one or more items remaining in the test vault
        - the orignal number items is greater than the number of remaining items
    """
    tags = ["tag_1"]
    vault_name = "Test Data 3"

    all_items_1 = signed_in_op.item_list(vault=vault_name)
    assert len(all_items_1)

    signed_in_op.item_delete_multiple(vault_name, tags=tags)
    all_items_2 = signed_in_op.item_list(vault=vault_name)

    assert len(all_items_2)
    assert len(all_items_1) > len(all_items_2)


@pytest.mark.usefixtures("setup_stateful_item_delete_multiple")
def test_item_delete_categories_multiple_01(signed_in_op: OP):
    """
    A test to delete all items of a certain category and verify none of those items
    remain afterwards

    retrieve list of all items in the test vault of category 'password'
    delete all items of category 'password'
    retrieve a second list of all items in the test vault of category 'password'

    verify:
        - before deleting, there are one or more password items in the list
        - after deleting, there are no password items
    """
    categories = ["password"]
    vault_name = "Test Data 3"

    orig_item_list = signed_in_op.item_list(
        vault=vault_name, categories=categories)
    assert len(orig_item_list)

    signed_in_op.item_delete_multiple(vault_name, categories=categories)
    new_item_list = signed_in_op.item_list(
        vault=vault_name, categories=categories)

    assert len(new_item_list) == 0


@pytest.mark.usefixtures("setup_stateful_item_delete_multiple_title_glob")
def test_item_delete_title_glob_multiple_01(signed_in_op: OP):
    """
    A test to delete all items matching a title glob pattern and verify none of those items
    remain afterwards

    retrieve list of all items in the test vault matching pattern 'Example Login Item *2'
    delete all items matching the glob pattern
    retrieve a second list of all items in the test vault matching the glob pattern

    verify:
        - before deleting, there are one or more items in the list
        - after deleting, there are no items matching the glob pattern
    """
    title_glob = "Example Login Item *2"
    # title_glob_alt = "Example Login Item *3"

    vault_name = "Test Data 3"

    orig_item_list = signed_in_op.item_list(
        vault=vault_name, title_glob=title_glob)
    assert len(orig_item_list)

    signed_in_op.item_delete_multiple(vault_name, title_glob=title_glob)
    new_item_list = signed_in_op.item_list(
        vault=vault_name, title_glob=title_glob)

    assert len(new_item_list) == 0


@pytest.mark.usefixtures("setup_stateful_item_delete_multiple_title_glob")
def test_item_delete_title_glob_multiple_02(signed_in_op: OP):
    """
    A test to delete all items matching a title glob pattern and verify items not matching
    the glob pattern are not deleted

    retrieve list of all items in the test vault matching alternate pattern 'Example Login Item *3'
    delete all items matching the glob pattern 'Example Login Item *2'
    retrieve a second list of all items in the test vault matching the alternate glob pattern

    verify:
        - before deleting, there are one or more items in the list matching the alternate glob pattern
        - after deleting, the same number of matching items remain
    """
    title_glob = "Example Login Item *2"
    title_glob_alt = "Example Login Item *3"

    vault_name = "Test Data 3"

    orig_item_list = signed_in_op.item_list(
        vault=vault_name, title_glob=title_glob_alt)
    assert len(orig_item_list) == 4

    signed_in_op.item_delete_multiple(vault_name, title_glob=title_glob)
    new_item_list = signed_in_op.item_list(
        vault=vault_name, title_glob=title_glob_alt)

    assert len(new_item_list) == len(orig_item_list)


def test_item_delete_multiple_non_existent_vault_01(signed_in_op: OP):
    vault = "Invalid Vault"
    with pytest.raises(OPItemDeleteMultipleException):
        signed_in_op.item_delete_multiple(vault)



from pathlib import Path

try:
    from importlib.resources import files as pkgfiles  # noqa: F401
except ImportError:  # pragma: no coverage
    from importlib_resources import files as pkgfiles  # noqa: F401


def data_location_as_path(package, sub_path) -> Path:
    """
    Convenience function to get a Path object from a package and a subpath
    within the package
    """
    # joinpath() returns a Traversable
    # to turn it into Path we first turn it into a string
    if not isinstance(sub_path, str):
        sub_path = str(sub_path)  # pragma: no coverage
    _data_path = pkgfiles(package).joinpath(sub_path)
    _data_path_str = str(_data_path)
    data_path = Path(_data_path_str)
    return data_path



import os
from typing import List, Union

from .json import safe_unjson
from .string import RedactedString

USER_UUID_UNMASK_ENV_VAR = "PYOP_UNMASK_USER_UUID"
ACCT_UUID_UNMASK_ENV_VAR = "PYOP_UNMASK_ACCOUNT_UUID"


class OPAccount(dict):

    # portion of unique IDs to leave unmasked
    UNMASK_LEN = 5

    def __init__(self, account_dict_or_json):
        account_dict = safe_unjson(account_dict_or_json)
        super().__init__(account_dict)

    @property
    def url(self) -> str:
        return self["url"]

    @property
    def email(self) -> str:
        return self["email"]

    @property
    def user_uuid(self) -> str:
        return self["user_uuid"]

    @property
    def account_uuid(self) -> str:
        return self["account_uuid"]

    @property
    def shorthand(self) -> Union[str, None]:
        return self.get("shorthand")

    # TODO: Not sure what to do here
    # the 'op whoami' dicts when signed in normally
    # vs. when using a service account
    # are really different
    # but neither self-describe (e.g. with an account type or similar)
    # so for now, just check if this has a ServiceAccountType key
    def is_service_account(self) -> bool:
        svc_acct = False
        if self.get("ServiceAccountType"):
            svc_acct = True  # pragma: no coverage
        return svc_acct

    @property
    def sanitized_account_uuid(self) -> str:
        _uuid = self.account_uuid
        if os.environ.get(ACCT_UUID_UNMASK_ENV_VAR, "0") != "1":
            _uuid = RedactedString(_uuid)
        return _uuid

    @property
    def sanitized_user_uuid(self) -> str:
        _uuid = self.user_uuid
        if os.environ.get(USER_UUID_UNMASK_ENV_VAR, "0") != "1":
            _uuid = RedactedString(_uuid)
        return _uuid


class OPAccountList(List[OPAccount]):

    def __init__(self, account_list_or_json):
        super().__init__()
        account_list = safe_unjson(account_list_or_json)
        for account_dict in account_list:
            op_account = OPAccount(account_dict)
            self.append(op_account)

    def account_for_identifier(self, account_id: str):
        acct = None
        for account in self:
            if account_id in [account.account_uuid, account.user_uuid, account.shorthand, account.email, account.url]:
                acct = account
                break
        return acct

    def user_id_for_account_identifier(self, account_id: str):
        user_id = None
        acct = self.account_for_identifier(account_id)
        if acct:
            user_id = acct.user_uuid
        return user_id



class RedactedString(str):
    """
    A string subclass that automatically self-redacts when printed, logged or passed to str()

    A portion at the end of the string will be unmasked by default. The unmasked lenght is configurable,
    per the following conditions:

    - The default unmask length is 5 characters
    - The maximum unmasked portion is 20% of the overall length
    - Strings fewer than 5 characters will remain completely masked
    """
    # how much to leave unmasked at the end
    UNMASK_LEN = 5
    UNMASK_LEN_MAX_FRACTION = 0.2
    MIN_STR_LEN = UNMASK_LEN

    _orig_string: str
    _redacted_string: str

    def __new__(cls, string, *args, unmask_len=-1, **kwargs):
        """
        Create a new RedactedString object

        Parameters
        ----------
        string : str
            The unredacted string
        unmask_len : int, optional
            The amount of the end of the string to leave unmasked, by default -1
            Notes:
                A maximum of 20% of the string length may be unmasked
                Strings of 5 characters or fewer will remain completely masked

        """
        if len(string) <= cls.MIN_STR_LEN:
            unmask_len = 0
        elif unmask_len < 0:
            unmask_len = cls.UNMASK_LEN

        max_fraction = cls.UNMASK_LEN_MAX_FRACTION * len(string)

        if unmask_len > max_fraction:
            unmask_len = int(max_fraction)

        orig_string = string
        if isinstance(string, cls):
            orig_string = string._orig_string
        obj = str.__new__(cls, orig_string)
        obj._orig_string = orig_string
        obj._redacted_string = cls._redacted(orig_string, unmask_len)
        return obj

    def __str__(self):
        return self._redacted_string

    def __iadd__(self, other):
        # since strings are immutable,
        # when you call mystr += "extra"
        # the string isn't modified
        # you actually get a new string object
        # so we can make this work just by calling __add__()
        return self.__add__(other)

    def __add__(self, other):
        if isinstance(other, RedactedString):
            other = other._orig_string
        new_str = RedactedString(self._orig_string + other)
        return new_str

    @property
    def original(self) -> str:
        return self._orig_string

    @classmethod
    def _redacted(cls, orig_string, unmask_len) -> str:
        # This aims to turn:
        # 5GHHPJK5HZC5BAT7WDUXW57G44 into
        # *********************57G44

        # get the length of the portion we're trying to mask
        star_count = len(orig_string) - unmask_len

        # make the star mask
        stars = "*" * star_count
        # get the bit we want to leave unmasked
        unmasked_end = orig_string[star_count:]

        # join the mask with the end part we're leaving unmasked
        redacted_string = stars + unmasked_end
        return redacted_string



"""
Description: A module that maps methods to to `op` commands and subcommands
"""
import enum
import logging
from os import environ
from typing import Dict, Mapping, Optional, Union

from ._op_cli_argv import _OPArgv
from ._op_cli_config import OPCLIConfig
from ._py_op_cli import _OPCLIExecute
from ._svc_account import (
    SVC_ACCT_CMD_NOT_SUPPORTED,
    SVC_ACCT_INCOMPAT_OPTIONS,
    SVC_ACCT_SUPPORTED,
    OPSvcAcctCommandNotSupportedException
)
from .account import OPAccount, OPAccountList
from .op_cli_version import (
    DOCUMENT_BYTES_BUG_VERSION,
    MINIMUM_SERVICE_ACCOUNT_VERSION,
    OPCLIVersion
)
from .py_op_exceptions import (
    OPAuthenticationException,
    OPCmdFailedException,
    OPCmdMalformedSvcAcctTokenException,
    OPDocumentDeleteException,
    OPDocumentGetException,
    OPGroupGetException,
    OPGroupListException,
    OPItemCreateException,
    OPItemDeleteException,
    OPItemGetException,
    OPItemListException,
    OPSigninException,
    OPUnknownAccountException,
    OPUserGetException,
    OPUserListException,
    OPVaultGetException,
    OPVaultListException,
    OPWhoAmiException
)


class ExistingAuthEnum(enum.IntEnum):
    IGNORE = 0     # Don't check if there's an existing authentication, peform new one regardless
    AVAILABLE = 1  # Check if there's an existing authentication, and if not, perform authentication
    REQUIRED = 2   # Check if there's an existing authentication, failing otherwise


EXISTING_AUTH_IGNORE = ExistingAuthEnum.IGNORE
EXISTING_AUTH_AVAIL = ExistingAuthEnum.AVAILABLE
EXISTING_AUTH_REQD = ExistingAuthEnum.REQUIRED


class _OPCommandInterface(_OPCLIExecute):
    """
    A class that directly maps methods to `op` commands
    & subcommands.
    No convenience methods are provided.
    No responses are parsed.
    """
    NOT_SIGNED_IN_TEXT = "not currently signed in"
    NO_ACTIVE_SESSION_FOUND_TEXT = "no active session found for account"
    NO_SESSION_TOKEN_FOUND_TEXT = "could not find session token for account"
    ACCT_IS_NOT_SIGNED_IN_TEXT = "account is not signed in"
    MALFORMED_SVC_ACCT_TEXT = "failed to DecodeSACCredentials"

    OP_SVC_ACCOUNT_ENV_VAR = "OP_SERVICE_ACCOUNT_TOKEN"
    OP_PATH = 'op'  # let subprocess find 'op' in the system path

    def __init__(self,
                 account: str = None,
                 password: str = None,
                 existing_auth: ExistingAuthEnum = EXISTING_AUTH_IGNORE,
                 vault: str = None,
                 password_prompt: bool = True,
                 op_path: str = OP_PATH,
                 logger: logging.Logger = None):
        """
        Constructor to authenticate or verify existing authentication to `op`
        """
        super().__init__()
        if not logger:
            logger = logging.getLogger(self.__class__.__name__)
            logger.setLevel(logging.INFO)

        self.vault = vault
        self.logger = logger
        self.op_path = op_path
        self._account_identifier = account
        self._signed_in_account: OPAccount = None

        self._cli_version: OPCLIVersion
        self._op_config: OPCLIConfig = None
        self._account_list: OPAccountList = None
        self._uses_bio: bool = False
        self._sess_var: str = None
        # gathering facts will attempt to set the above instance variables
        # that got initialized to None or False
        self._gather_facts()

        # Coerce existing_auth to an Enum in case it was passed in as a legacy bool
        # False -> ExistingAuthFlag.NONE, True -> ExistingAuthFlag.AVAILABLE
        existing_auth = ExistingAuthEnum(existing_auth)

        # part of exception message in case incompatible authentication parameters
        # are provided
        auth_pref_source = "preference passed as argument"
        if self.svc_account_env_var_set():
            # - 'op' won't prompt for authentication of OP_SERVICE_ACCOUNT_TOKEN is set
            # - Even if we explicitly call signin and succeed, it ignores that authentication
            # so we need to suppress any path that tries to or expects to authenticate
            if self._cli_version < MINIMUM_SERVICE_ACCOUNT_VERSION:
                raise OPAuthenticationException(
                    f"Version {self._cli_version} not supported with service accounts. Minimum version: {MINIMUM_SERVICE_ACCOUNT_VERSION}")
            if existing_auth != EXISTING_AUTH_REQD:
                auth_pref_source = "preference upgraded due to service account environment variable"
                self.logger.info(
                    f"{self.OP_SVC_ACCOUNT_ENV_VAR} was set. Upgrading existing auth flag to REQUIRED")
                existing_auth = EXISTING_AUTH_REQD

        if existing_auth == EXISTING_AUTH_REQD and password is not None:
            # if a password is passed in but existing_auth is required, caller may be confused:
            # - intentionally passed in incompatible options
            # - possibly has OP_SERVICE_ACCOUNT_TOKEN set accidentally
            msg = f"Password argument passed but EXISTING_AUTH_REQD flag is set. flag source: {auth_pref_source}"
            self.logger.error(msg)
            raise OPAuthenticationException(msg)

        # So far everything above has been fairly lightweight, with no remote
        # contact to the 1Password account.
        # The next steps will attempt to talk to the 1Password account, and
        # failing that, may attempt to authenticate to the 1Password account
        account_obj, token = self._new_or_existing_signin(
            existing_auth, password, password_prompt)
        self._signed_in_account = account_obj
        self._token = token
        if self._signed_in_account.is_service_account():
            self.logger.debug("Signed in as a service account")
        else:
            self.logger.debug(
                f"Signed in as User ID: {self._signed_in_account.sanitized_user_uuid}")
        # export OP_SESSION_<use_id>
        if self.token:
            if not self._account_identifier:
                # if we weren't provided an account identifier,
                # we can now get it from the signed-in account object
                # and compute the session environment variable name
                self._account_identifier = account_obj.user_uuid
                self._sess_var = self._compute_session_var_name()
            environ[self._sess_var] = self.token

    @property
    def token(self) -> str:
        return self._token

    @property
    def session_var(self) -> str:
        return self._sess_var

    @classmethod
    def svc_account_env_var_set(cls):
        svc_acct_set = False
        if environ.get(cls.OP_SVC_ACCOUNT_ENV_VAR):
            svc_acct_set = True
        return svc_acct_set

    @classmethod
    def uses_biometric(cls, op_path: str = "op", encoding: str = "utf-8", account_list: OPAccountList = None):
        uses_bio = True
        # We can run 'op account list', which doesn't require talking (or authenticating)
        # to the 1Password account
        # if biometric is enabled, there will be no account shorthands in the output
        # if there are account shorthands, biometric is not enabled
        if account_list is None:
            account_list = cls._get_account_list(op_path, decode=encoding)
        acct: OPAccount
        for acct in account_list:
            if not acct.shorthand:
                continue
            # There is at least one account_shorthand found in `op account list`
            uses_bio = False
            break
        return uses_bio

    def _gather_facts(self):
        self._op_config = OPCLIConfig()
        self._cli_version = self._get_cli_version(self.op_path)
        self._account_list = self._get_account_list(self.op_path)
        self._uses_bio = self.uses_biometric(
            op_path=self.op_path, account_list=self._account_list)
        self._account_identifier = self._normalize_account_id()
        self._sess_var = self._compute_session_var_name()

    def _get_cli_version(self, op_path: str) -> OPCLIVersion:
        argv = _OPArgv.cli_version_argv(op_path)
        output = self._run(argv, capture_stdout=True, decode="utf-8")
        output = output.rstrip()
        cli_version = OPCLIVersion(output)
        return cli_version

    @classmethod
    def _get_account_list(cls, op_path, decode="utf-8") -> OPAccountList:
        account_list_json = cls._signed_in_accounts(op_path, decode=decode)
        account_list = OPAccountList(account_list_json)
        return account_list

    def _normalize_account_id(self):
        # we need to turn whatever we were given into a User ID
        # we could have been given any of the following for account ID:
        # - None
        # - account shorthand
        # - account URL
        # - account UUID
        # - user UUID
        # see: 'op --help' for '--account' global flag
        user_uuid = None
        if not self._uses_bio:
            if self._account_identifier:
                # if we were given an specific account look up user ID for that
                user_uuid = self._op_config.uuid_for_account(
                    self._account_identifier)
            else:
                # else try to look up User ID for latest sign-in
                # (may come back None)
                user_uuid = self._op_config.latest_signin_uuid
        elif self._account_identifier:
            user_uuid = self._account_list.user_id_for_account_identifier(
                self._account_identifier)

        if self._account_identifier and not user_uuid:
            raise OPUnknownAccountException(
                f"No account found for identifier: {self._account_identifier}")

        return user_uuid

    def _compute_session_var_name(self):
        sess_var_name = self._sess_var
        # normalize account identifier before calling this
        user_uuid = self._account_identifier

        # we can only use a session variable if:
        #   - biometric isn't enabled, and
        #   - we have an account shorthand

        if user_uuid and not self._uses_bio:
            sess_var_name = 'OP_SESSION_{}'.format(user_uuid)
        return sess_var_name

    def _new_or_existing_signin(self, existing_auth: ExistingAuthEnum, password: str, password_prompt: bool):
        token = None
        account = None

        # Don't attempt to verify sign-in unless caller told us to
        # otherwise it will trigger a prompt, either GUI biometric prompt,
        # or interactive console prompt
        if existing_auth in [EXISTING_AUTH_AVAIL, EXISTING_AUTH_REQD]:
            account = self._verify_signin()

        if account:
            token = self._get_existing_token(account)
        else:
            if existing_auth == EXISTING_AUTH_REQD:
                # we were told to only use existing authentication but verificaiton failed
                # this is a hard error
                raise OPAuthenticationException(
                    "Existing authentication specified as required, but could not be verified.")

            # we couldn't verify being signed in (or weren't told to try)
            # let's try a normal sign-in
            # _do_normal_signin() will raise OPAuthenticationException if
            # _uses_bio is false, no password given, and password prompt not allowed
            token = self._do_normal_signin(password, password_prompt)
            account = self._verify_signin(token=token)
        return (account, token)

    @classmethod
    def _auth_expired(cls, op_path, account):
        # this is a test to see if the previously valid authentication
        # is still valid, with the assumption that if it is not, then it
        # has expired
        # it is primarily for the following two purposes
        # - avoid triggering an interactive prompt or GUI dialogue (if undesired) and hanging indefinitely
        # - being able to raise OPAuthenticationException to the caller rather than a generic
        #   "command failed" exception
        expired = False

        try:
            cls._whoami(op_path, account=account)
        except OPWhoAmiException:  # pragma: no cover
            expired = True

        return expired

    def _verify_signin(self, token: str = None):
        env: Mapping
        account = None
        # copy os.environ to a dict since its type is incompatible
        # with Dict[str, str], as reported by mypy
        env = dict(environ)
        if token and self._sess_var:
            # we need to pass an environment dictionary to subprocess.run() that contains
            # the session token we're trying to verify
            #
            # Above we made a copy of os.environ, so we're not modifying it at this point
            # we'll offically set it once we know it works
            env[self._sess_var] = token

        # this step actually talks to the 1Password account
        # it uses "op whoami" which is a very non-intrusive
        # query that will fail without authentication
        try:
            account = self._whoami(self.op_path, env=env)
        except OPWhoAmiException as ocfe:
            # scrape error message about not being signed in

            fragments = [self.NO_ACTIVE_SESSION_FOUND_TEXT,
                         self.NOT_SIGNED_IN_TEXT,
                         self.NO_SESSION_TOKEN_FOUND_TEXT,
                         self.ACCT_IS_NOT_SIGNED_IN_TEXT]
            unknown_err = True
            for frag in fragments:
                if frag in ocfe.err_output:
                    unknown_err = False
                    break
            # there was a different error so raise the exception
            if unknown_err:  # pragma: no cover
                raise ocfe

        return account

    def _do_normal_signin(self, password: str, password_prompt: bool) -> Union[str, None]:
        # normalize empty string to None, otherwise use password as given
        password = None if password == "" else password

        # there are three things that can be used for a *new* authentication:
        # - biometric
        # - password
        # - allow 'op' to interactively prompt
        # NOTE: we need to keep this check as close to the call to _run_signin() as possible
        #       We can't tell 'op' not to prompt. We just have to not run it if we're not prompting
        if not self._uses_bio and not password and not password_prompt:
            # - we weren't provided a password, and
            # - we were told not to let 'op' prompt for a password, and
            # - biometric is not enabled
            # so we can't sign in
            raise OPAuthenticationException(
                "No existing session and no password provided.")
        signin_argv = _OPArgv.normal_signin_argv(
            self.op_path, account=self._account_identifier)

        token = self._run_signin(signin_argv, password=password)
        if self._uses_bio:
            # Set to None vs empty string
            token = None

        return token

    def _get_existing_token(self, account: OPAccount):
        if self._sess_var:
            token = environ.get(self._sess_var)
        else:
            token = None
        return token

    def _run_signin(self, argv, password=None):
        try:
            output = self._run(argv, capture_stdout=True,
                               input_string=password, decode="utf-8")
        except OPCmdFailedException as ocfe:
            raise OPSigninException.from_opexception(ocfe) from ocfe

        return output

    @classmethod
    def _run_with_auth_check(cls,
                             op_path: str,
                             account: str,
                             argv: _OPArgv,
                             capture_stdout: bool = False,
                             input_string: str = None,
                             decode: str = None,
                             env: Mapping = environ):
        # this somewhat of a hack to detect if authentication has expired
        # so that we can raise OPAuthenticationException rather than the generic OPCmdFailedException
        # under the hood, it calls 'whoami' which will fail if not authenticated
        #
        # We need to remove this as soon as possible if a better way becomes available
        # among the problems:
        # - this method is racey, since authentication may expire between the check and the
        #   operation
        # - this adds roughly 20% overhead (as measured by the full suite of pytest tests)
        if cls._auth_expired(op_path, account):
            raise OPAuthenticationException(
                "Authentication has expired")  # pragma: no cover

        if cls.svc_account_env_var_set():
            err_msg = None
            supported = argv.svc_account_supported()
            if supported.code in [SVC_ACCT_INCOMPAT_OPTIONS, SVC_ACCT_CMD_NOT_SUPPORTED]:
                err_msg = supported.msg
            elif supported.code == SVC_ACCT_SUPPORTED:
                cls.logger.debug("Command supported with service accounts")
            else:
                raise Exception(  # pragma: no cover
                    f"Unknown service account support code {supported.code}")

            if err_msg:
                if cls._should_log_op_errors():
                    cls.logger.error(err_msg)
                raise OPSvcAcctCommandNotSupportedException(err_msg)

        return cls._run(argv,
                        capture_stdout=capture_stdout,
                        input_string=input_string,
                        decode=decode,
                        env=env)

    @classmethod
    def _account_list_argv(cls, op_path="op", encoding="utf-8"):
        argv = _OPArgv.account_list_argv(op_path, encoding=encoding)
        return argv

    def _item_get_argv(self, item_name_or_id, vault=None, fields=None, include_archive=False):
        vault_arg = vault if vault else self.vault

        lookup_argv = _OPArgv.item_get_argv(
            self.op_path, item_name_or_id, vault=vault_arg, fields=fields, include_archive=include_archive)
        return lookup_argv

    def _item_delete_argv(self, item_name_or_id, vault=None, archive=False):
        vault_arg = vault if vault else self.vault

        delete_argv = _OPArgv.item_delete_argv(
            self.op_path, item_name_or_id, vault=vault_arg, archive=archive)
        return delete_argv

    def _item_get_totp_argv(self, item_name_or_id, vault=None):
        vault_arg = vault if vault else self.vault

        lookup_argv = _OPArgv.item_get_totp_argv(
            self.op_path, item_name_or_id, vault=vault_arg)
        return lookup_argv

    def _document_get_argv(self,
                           document_name_or_id: str,
                           vault: Optional[str] = None,
                           include_archive: Optional[bool] = False):
        vault_arg = vault if vault else self.vault
        document_get_argv = _OPArgv.document_get_argv(self.op_path,
                                                      document_name_or_id,
                                                      vault=vault_arg,
                                                      include_archive=include_archive)
        print(document_get_argv)

        return document_get_argv

    def _document_delete_argv(self, document_name_or_id: str, vault: Optional[str] = None, archive=False):
        vault_arg = vault if vault else self.vault

        document_delete_argv = _OPArgv.document_delete_argv(
            self.op_path, document_name_or_id, vault=vault_arg, archive=archive)

        return document_delete_argv

    def _user_get_argv(self, user_name_or_id: str):
        get_user_argv = _OPArgv.user_get_argv(self.op_path, user_name_or_id)
        return get_user_argv

    def _user_list_argv(self, group_name_or_id=None, vault=None):
        user_list_argv = _OPArgv.user_list_argv(
            self.op_path, group_name_or_id=group_name_or_id, vault=vault)
        return user_list_argv

    def _group_get_argv(self, group_name_or_id: str):
        group_get_argv = _OPArgv.group_get_argv(
            self.op_path, group_name_or_id)
        return group_get_argv

    def _group_list_argv(self, user_name_or_id=None, vault=None):
        group_list_argv = _OPArgv.group_list_argv(
            self.op_path, user_name_or_id=user_name_or_id, vault=vault)
        return group_list_argv

    def _vault_get_argv(self, vault_name_or_id: str):

        get_vault_argv = _OPArgv.vault_get_argv(
            self.op_path, vault_name_or_id)
        return get_vault_argv

    def _vault_list_argv(self, group_name_or_id=None, user_name_or_id=None):
        vault_list_argv = _OPArgv.vault_list_argv(
            self.op_path, group_name_or_id=group_name_or_id, user_name_or_id=user_name_or_id)
        return vault_list_argv

    @classmethod
    def _whoami(cls, op_path, env: Dict[str, str] = None, account: str = None) -> OPAccount:
        if not env:
            env = dict(environ)
        argv = _OPArgv.whoami_argv(op_path, account=account)
        try:
            account_json = cls._run(
                argv, capture_stdout=True, decode="utf-8", env=env)
        except OPCmdFailedException as ocfe:
            # scrape error message about not being signed in

            if cls.MALFORMED_SVC_ACCT_TEXT in ocfe.err_output:
                # OP_SERVICE_ACCOUNT_TOKEN got set to something malformed
                # so raise a specific exception for that
                raise OPCmdMalformedSvcAcctTokenException.from_opexception(
                    ocfe)  # pragma: no cover
                # Although we could simulate this for testing, the tests
                # wouldn't be meaningful, because they wouldn't be tied to
                # an actual malformed token
                # disabling testing coverage
            else:
                raise OPWhoAmiException.from_opexception(ocfe)

        account_obj = OPAccount(account_json)
        return account_obj

    def _item_get(self, item_name_or_id, vault=None, fields=None, include_archive=False, decode="utf-8"):
        get_item_argv = self._item_get_argv(
            item_name_or_id, vault=vault, fields=fields, include_archive=include_archive)
        try:
            output = self._run_with_auth_check(
                self.op_path, self._account_identifier, get_item_argv, capture_stdout=True, decode=decode)
        except OPCmdFailedException as ocfe:
            raise OPItemGetException.from_opexception(ocfe) from ocfe

        return output

    def _item_delete(self, item_name_or_id, vault=None, archive=False, decode="utf-8"):
        item_delete_argv = self._item_delete_argv(
            item_name_or_id, vault=vault, archive=archive)
        try:
            # 'op item delete' doesn't have any output if successful
            # if it fails, stderr will be in the exception object
            self._run_with_auth_check(
                self.op_path, self._account_identifier, item_delete_argv, decode=decode)
        except OPCmdFailedException as ocfe:
            raise OPItemDeleteException.from_opexception(ocfe)

        return

    def _item_delete_multiple(self, batch_json, vault, archive=False):
        # op item delete takes '-' for the item to delete if objects are
        # provided over stdin
        item_id = "-"
        item_delete_argv = self._item_delete_argv(
            item_id, vault=vault, archive=archive)
        try:
            # 'op item delete' doesn't have any output if successful
            # if it fails, stderr will be in the exception object
            self._run_with_auth_check(
                self.op_path, self._account_identifier, item_delete_argv, input_string=batch_json)
        except OPCmdFailedException as ocfe:
            raise OPItemDeleteException.from_opexception(ocfe)

        return

    def _item_get_totp(self, item_name_or_id, vault=None, decode="utf-8"):
        item_get_totp_argv = self._item_get_totp_argv(
            item_name_or_id, vault=vault)
        try:
            output = self._run_with_auth_check(self.op_path, self._account_identifier,
                                               item_get_totp_argv, capture_stdout=True, decode=decode)
        except OPCmdFailedException as ocfe:
            raise OPItemGetException.from_opexception(ocfe) from ocfe

        return output

    def _document_get(self,
                      document_name_or_id: str,
                      vault: Optional[str] = None,
                      include_archive: Optional[bool] = False):
        """
        Download a document object from a 1Password vault by name or UUID.

        Arguments:
            - 'item_name_or_id': The item to look up
        Raises:
            - OPDocumentGetException if the lookup fails for any reason.
            - OPNotFoundException if the 1Password command can't be found.
        Returns:
            - Bytes: document bytes
        """

        get_document_argv = self._document_get_argv(
            document_name_or_id, vault=vault, include_archive=include_archive)

        try:
            document_bytes = self._run_with_auth_check(
                self.op_path, self._account_identifier, get_document_argv, capture_stdout=True)
        except OPCmdFailedException as ocfe:
            raise OPDocumentGetException.from_opexception(ocfe) from ocfe

        if self._cli_version <= DOCUMENT_BYTES_BUG_VERSION:  # pragma: no cover
            # op v2.x appends an erroneous \x0a ('\n') byte to document bytes
            # trim it off if its present
            if document_bytes[-1] == 0x0a:
                document_bytes = document_bytes[:-1]

        return document_bytes

    def _document_delete(self, document_name_or_id: str, vault: Optional[str] = None, archive=False):

        document_delete_argv = self._document_delete_argv(
            document_name_or_id, vault=vault, archive=archive)
        try:
            # 'op document delete' doesn't have any output if successful
            # if it fails, stderr will be in the exception object
            self._run_with_auth_check(
                self.op_path, self._account_identifier, document_delete_argv)
        except OPCmdFailedException as ocfe:
            raise OPDocumentDeleteException.from_opexception(ocfe)

        return

    @classmethod
    def _signed_in_accounts(cls, op_path, decode="utf-8"):
        account_list_argv = cls._account_list_argv(op_path, encoding=decode)
        output = cls._run(account_list_argv,
                          capture_stdout=True, decode=decode)
        return output

    def _user_get(self, user_name_or_id: str, decode: str = "utf-8") -> str:
        get_user_argv = self._user_get_argv(user_name_or_id)
        try:
            output = self._run_with_auth_check(self.op_path, self._account_identifier,
                                               get_user_argv, capture_stdout=True, decode=decode)
        except OPCmdFailedException as ocfe:
            raise OPUserGetException.from_opexception(ocfe) from ocfe
        return output

    def _user_list(self, group_name_or_id=None, vault=None, decode: str = "utf-8") -> str:
        user_list_argv = self._user_list_argv(
            group_name_or_id=group_name_or_id, vault=vault)
        try:
            output = self._run_with_auth_check(self.op_path, self._account_identifier,
                                               user_list_argv, capture_stdout=True, decode=decode)
        except OPCmdFailedException as ocfe:
            raise OPUserListException.from_opexception(ocfe)
        return output

    def _group_get(self, group_name_or_id: str, decode: str = "utf-8") -> str:
        group_get_argv = self._group_get_argv(group_name_or_id)
        try:
            output = self._run_with_auth_check(self.op_path, self._account_identifier,
                                               group_get_argv, capture_stdout=True, decode=decode)
        except OPCmdFailedException as ocfe:
            raise OPGroupGetException.from_opexception(ocfe) from ocfe
        return output

    def _group_list(self, user_name_or_id=None, vault=None, decode: str = "utf-8") -> str:
        group_list_argv = self._group_list_argv(
            user_name_or_id=user_name_or_id, vault=vault)
        try:
            output = self._run_with_auth_check(self.op_path, self._account_identifier,
                                               group_list_argv, capture_stdout=True, decode=decode)
        except OPCmdFailedException as ocfe:
            raise OPGroupListException.from_opexception(ocfe)
        return output

    def _vault_get(self, vault_name_or_id: str, decode: str = "utf-8") -> str:
        vault_get_argv = self._vault_get_argv(vault_name_or_id)
        try:
            output = self._run_with_auth_check(self.op_path, self._account_identifier,
                                               vault_get_argv, capture_stdout=True, decode=decode)
        except OPCmdFailedException as ocfe:
            raise OPVaultGetException.from_opexception(ocfe)
        return output

    def _vault_list(self, group_name_or_id=None, user_name_or_id=None, decode="utf-8") -> str:
        vault_list_argv = self._vault_list_argv(
            group_name_or_id=group_name_or_id, user_name_or_id=user_name_or_id)
        try:
            output = self._run_with_auth_check(self.op_path, self._account_identifier,
                                               vault_list_argv, capture_stdout=True, decode=decode)
        except OPCmdFailedException as ocfe:
            raise OPVaultListException.from_opexception(ocfe)
        return output

    def _signout(self, account, session, forget=False):   # pragma: no cover
        if forget and self._uses_bio:
            self.logger.warn(
                "Biometric is enabled. 'forget' operation will have no effect.")
        argv = _OPArgv.signout_argv(
            self.op_path, account, session, forget=forget, uses_bio=self._uses_bio)
        self._run(argv)

    @classmethod
    def _account_forget(cls, account: str, op_path=None):   # pragma: no cover
        if not op_path:
            op_path = cls.OP_PATH
        argv = _OPArgv.account_forget_argv(op_path, account)
        cls._run(argv)

    def _item_list_argv(self, categories=[], include_archive=False, tags=[], vault=None):
        vault_arg = vault if vault else self.vault
        list_items_argv = _OPArgv.item_list_argv(self.op_path,
                                                 categories=categories, include_archive=include_archive, tags=tags, vault=vault_arg)
        return list_items_argv

    def _item_list(self, categories=[], include_archive=False, tags=[], vault=None, decode="utf-8"):
        argv = self._item_list_argv(
            categories=categories, include_archive=include_archive, tags=tags, vault=vault)
        try:
            output = self._run_with_auth_check(
                self.op_path, self._account_identifier, argv, capture_stdout=True, decode=decode)
        except OPCmdFailedException as e:
            raise OPItemListException.from_opexception(e)
        return output

    def _item_create_argv(self, item, password_recipe, vault):
        vault_arg = vault if vault else self.vault
        create_item_argv = _OPArgv.item_create_argv(
            self.op_path, item, password_recipe=password_recipe, vault=vault_arg
        )
        return create_item_argv

    def _item_create(self, item, vault, password_recipe, decode="utf-8"):
        argv = self._item_create_argv(item, password_recipe, vault)
        try:
            output = self._run_with_auth_check(
                self.op_path, self._account_identifier, argv, capture_stdout=True, decode=decode)
        except OPCmdFailedException as e:
            raise OPItemCreateException.from_opexception(e)

        return output



from typing import List, Optional, Union

from .._item_descriptor_registry import op_register_item_descriptor_type
from .._item_type_registry import op_register_item_type
from .._new_item import OPNewItemMixin
from ..fields_sections._new_fields import (
    OPNewPasswordField,
    OPNewUsernameField
)
from ..fields_sections.item_field_base import OPItemField
from ..fields_sections.item_section import OPSection
from ._item_base import OPAbstractItem
from ._item_descriptor_base import OPAbstractItemDescriptor


class OPNewLoginItemURLException(Exception):
    pass


@op_register_item_descriptor_type
class OPLoginDescriptorItem(OPAbstractItemDescriptor):
    CATEGORY = "LOGIN"

    def __init__(self, item_dict):
        super().__init__(item_dict)


class OPLoginItemURL(dict):
    def __init__(self, url_dict):
        super().__init__(url_dict)

    @property
    def label(self) -> Optional[str]:
        return self.get("label")

    @property
    def primary(self) -> Optional[bool]:
        primary = self.get("primary", False)
        return primary

    @property
    def href(self) -> str:
        return self["href"]


class OPLoginItemNewURL(OPLoginItemURL):
    """
    A class to create a new URL dictionary for use with OPLoginItemTemplate
    """

    def __init__(self, url: str, label: str, primary: bool = False):
        """
        Create a new OPLoginItemNewURL object

        Parameters
        ----------
        url: str
            The URL for this object
        label: str
            The user-visible name of the URL entry
        primary: bool, optional
            Whether this URL should be flagged as primary. Defaults to False
            NOTE: There can only be one primary URL in a login item
        """
        url_dict = {
            "label": label,
            "primary": primary,
            "href": url
        }
        super().__init__(url_dict)


class OPLoginItemNewPrimaryURL(OPLoginItemNewURL):
    """
    A class to create a new primary URL dictionary for use with OPLoginItemTemplate
    """

    def __init__(self, url: str, label: str):
        super().__init__(url, label, primary=True)


@op_register_item_type
class OPLoginItem(OPAbstractItem):
    CATEGORY = "LOGIN"

    def __init__(self, item_dict_or_json):
        super().__init__(item_dict_or_json)
        urls = []
        primary_url = None
        url_list = self.get("urls", [])
        for url_dict in url_list:
            url = OPLoginItemURL(url_dict)
            if url.primary:
                primary_url = url
            urls.append(url)
        self._urls = urls
        self._primary_url = primary_url

    @property
    def username(self):
        username = self.field_value_by_id("username")
        return username

    @property
    def password(self):
        password = self.field_value_by_id("password")
        return password

    @property
    def urls(self) -> List[OPLoginItemURL]:
        return self._urls

    @property
    def primary_url(self) -> OPLoginItemURL:
        return self._primary_url


@op_register_item_type
class OPLoginItemRelaxedValidation(OPLoginItem):
    # see ITEM_VALIDATION.md
    _relaxed_validation = True


class OPLoginItemTemplate(OPNewItemMixin, OPLoginItem):
    """
    Class for creating a login item template that can be used to create a new login item in 1Password
    """

    PASSWORDS_SUPPORTED = True

    FIELD_ID_USERNAME = "username"
    FIELD_ID_PASSWORD = "password"
    DEFAULT_URL_LABEL = "website"

    def __init__(self,
                 title: str,
                 username: str,
                 password: Optional[str] = None,
                 url: Union[str, OPLoginItemURL, None] = None,
                 fields: List[OPItemField] = [],
                 sections: List[OPSection] = [],
                 tags: List[str] = []):
        """
        Create an OPLoginItemTemplate object that can be used to create a new login item entry

        Parameters
        ----------
        Title : str
            User viewable name of the login item to create
        username : str
            username string for the new login item
        password : str, optional
            password to set for this login item
        url: Union[str, OPLoginItemURL] = None
            If provided, set to the primary URL of the login item
            If URL is a str, it is converted to OPLoginItemURL, with label="website"
        fields: List[OPItemField]
            List of OPItemField objects to associate with the item.
            NOTE: If the fields are from an exisiting item, and the field IDs are UUIDs, the field IDs will be regenerated
        sections: List[OPSection]
            List of OPSection objects to associate with the item.
            NOTE: If the sections are from an exisiting item, and the section IDs are UUIDs, the section IDs will be regenerated
        tags: List[str], optional
            A list of tags to apply to the login item template
        Raises
        ------
        OPNewLoginItemURLException
            If an OPUrlItem object is provided to the constructor, and it's not flagged as a primary URL
        """
        if fields is None:  # pragma: no coverage
            fields = []
        else:
            fields = list(fields)

        if isinstance(url, str):
            url = OPLoginItemNewPrimaryURL(url, self.DEFAULT_URL_LABEL)

        username_field = OPNewUsernameField(
            self.FIELD_ID_USERNAME,
            username,
            field_id=self.FIELD_ID_USERNAME
        )

        password_field = OPNewPasswordField(
            self.FIELD_ID_PASSWORD,
            password,
            field_id=self.FIELD_ID_PASSWORD
        )

        fields.extend([username_field, password_field])
        urls = []
        if url:
            if not url.primary:
                raise OPNewLoginItemURLException("Sole URL must be primary")
            urls.append(url)
        extra_data = {"urls": urls}
        super().__init__(title, fields=fields, sections=sections,
                         tags=tags, extra_data=extra_data)

    def add_url(self, url: OPLoginItemURL):
        """
        Add a URL to this login item

        Parameters
        ----------
        url: OPLoginItemURL
            The URL to add to this login item

        Raises
        ------
        OPNewLoginItemURLException
            If the URL object is a primary URL, and there is already a primary URL
        """
        has_primary = False
        if self.primary_url:
            has_primary = True
        if url.primary and has_primary:
            raise OPNewLoginItemURLException("Can't add multiple primary URLs")

        self._urls.append(url)
        if url.primary:
            self._primary_url = url



from typing import List, Union

from .._item_descriptor_registry import op_register_item_descriptor_type
from .._item_type_registry import op_register_item_type
from .._new_item import OPNewItemMixin
from ..fields_sections._new_fields import (
    OPNewConcealedField,
    OPNewNetworkPortField,
    OPNewStringField
)
from ..fields_sections.item_field import OPItemField
from ..fields_sections.item_section import OPSection
from ._item_base import OPAbstractItem, OPFieldNotFoundException
from ._item_descriptor_base import OPAbstractItemDescriptor


@op_register_item_descriptor_type
class OPDatabaseItemDescriptor(OPAbstractItemDescriptor):
    CATEGORY = "DATABASE"

    def __init__(self, item_dict):
        super().__init__(item_dict)


@op_register_item_type
class OPDatabaseItem(OPAbstractItem):
    CATEGORY = "DATABASE"

    def __init__(self, item_dict):
        super().__init__(item_dict)

    @property
    def database_type(self) -> str:
        try:
            database_type = self.field_value_by_id("database_type")
        except OPFieldNotFoundException:  # pragma: no coverage
            database_type = None
        return database_type

    @property
    def type(self) -> str:
        # convenience accessor for "database_type"
        # label for "database_type" is "type"
        return self.database_type

    @property
    def hostname(self) -> str:
        try:
            hostname = self.field_value_by_id("hostname")
        except OPFieldNotFoundException:
            hostname = None
        return hostname

    @property
    def server(self) -> str:
        # convenience accessor for "hostname"
        # label for "hostname" is "server"
        return self.hostname

    @property
    def port(self) -> int:
        try:
            port = self.field_value_by_id("port")
        except OPFieldNotFoundException:
            port = None
        return port

    @property
    def database(self) -> Union[str, None]:
        try:
            database = self.field_value_by_id("database")
        except OPFieldNotFoundException:
            database = None
        return database

    @property
    def username(self) -> str:
        try:
            username = self.field_value_by_id("username")
        except OPFieldNotFoundException:
            username = None
        return username

    @property
    def password(self) -> str:
        try:
            password = self.field_value_by_id("password")
        except OPFieldNotFoundException:
            password = None
        return password

    @property
    def sid(self) -> Union[str, None]:
        try:
            sid = self.field_value_by_id("SID")
        except OPFieldNotFoundException:
            sid = None
        return sid

    @property
    def alias(self) -> Union[str, None]:
        try:
            alias = self.field_value_by_id("alias")
        except OPFieldNotFoundException:
            alias = None
        return alias

    @property
    def options(self) -> Union[str, None]:
        try:
            options = self.field_value_by_id("options")
        except OPFieldNotFoundException:
            options = None
        return options

    @property
    def connection_options(self) -> Union[str, None]:
        # convenience accessor for "options"
        # label for "options" field is "connection options"
        return self.options


@op_register_item_type
class OPDatabaseItemRelaxedValidation(OPDatabaseItem):
    # see ITEM_VALIDATION.md
    _relaxed_validation = True


class OPDatabaseItemTemplate(OPNewItemMixin, OPDatabaseItem):
    """
    Class for creating a database item template that can be used to create a new database item in 1Password
    """

    # Not setting PASSWORDS_SUPPORTED = True
    # because 'op' only supports password creation for LOGIN and PASSWORD
    # item types

    def __init__(self,
                 title: str,
                 database_type: str = None,
                 hostname: str = None,
                 port: Union[str, int] = None,
                 database: str = None,
                 username: str = None,
                 password: str = None,
                 sid: str = None,
                 alias: str = None,
                 options: str = None,
                 fields: List[OPItemField] = [],
                 sections: List[OPSection] = [],
                 tags: List[str] = []):
        """
        Create an OPDatabaseItemTemplate object that can be used to create a new database item entry

        Parameters
        ----------
        title : str
            User viewable name of the database item to create
        database_type : str, optional
            Value for the database type field, by default None
        hostname : str, optional
            Value for the hostname/server field, by default None
        port : Union[str, int], optional
            port value for the new database item, by default None
            Discussion:
                The port value may be either an integer or a string representation of an integer.

                If an integer is provided, it will be converted to a string via 'str(value)'

                If a string is provided, it will be validated by converting it to an integer via 'int(value, 0)'.
                If this fails, ValueError is raised.
                The string may be any valid base representation. E.g., 1234 or 0x1234. This representation will
                preserved.

                No attempt is made to ensure the value is in a reasonable range for what a port should be

        database : str, optional
            Value for the database field, by default None
        username : str, optional
            Value for the username field, by default None
        password : str, optional
            Value for the password field, by default None
        sid : str, optional
            Value for the SID field, by default None
        alias : str, optional
            Value for the alias field, by default None
        options : str, optional
            Value for the connection options field, by default None
        fields: List[OPItemField]
            List of OPItemField objects to associate with the item.
            NOTE: If the fields are from an exisiting item, and the field IDs are UUIDs, the field IDs will be regenerated
        sections: List[OPSection]
            List of OPSection objects to associate with the item.
            NOTE: If the sections are from an exisiting item, and the section IDs are UUIDs, the section IDs will be regenerated
        tags: List[str], optional
            A list of tags to apply to the login item template

        Raises
        ------
        ValueError
            If the port value fails validation as described above
        """

        if fields is None:  # pragma: no coverage
            fields = []
        else:
            fields = list(fields)

        if database_type:
            # label for database_type is "type"
            database_type_field = OPNewStringField(
                "type", database_type, field_id="database_type")
            fields.append(database_type_field)

        if hostname:
            # label for hostname field is "server"
            hostname_field = OPNewStringField(
                "server", hostname, field_id="hostname")
            fields.append(hostname_field)

        if port:
            port_field = OPNewNetworkPortField("port", port, field_id="port")
            fields.append(port_field)

        if sid:
            sid_field = OPNewStringField("SID", sid, field_id="sid")
            fields.append(sid_field)

        if database:
            # database name
            database_field = OPNewStringField(
                "database", database, field_id="database")
            fields.append(database_field)

        if username:
            # we don't use OPNewUsernameField because database username
            # doesn't declare a FIELD_PURPOSE = "USERNAME"
            username_field = OPNewStringField(
                "username", username, field_id="username")
            fields.append(username_field)

        if password:
            # we don't use OPNewPasswordField because database password
            # doesn't declare a FIELD_PURPOSE = "PASSWORD"
            password_field = OPNewConcealedField(
                "password", password, field_id="password")
            fields.append(password_field)

        if alias:
            alias_field = OPNewStringField("alias", alias, field_id="alias")
            fields.append(alias_field)

        if options:
            # label for options field is "database options"
            options_field = OPNewStringField(
                "database options", options, field_id="options")
            fields.append(options_field)

        super().__init__(title, fields, sections, tags)



from pyonepassword.op_cli_version import OPCLIVersion

VERSION_STRING_2_0_0 = "2.0.0"
VERSION_STRING_2_0_0_beta_01 = "2.0.0-beta.01"
VERSION_STRING_2_0_0_beta_14 = "2.0.0-beta.14"

VERSION_STRING_2_1_0 = "2.1.0"
VERSION_STRING_2_1_0_beta_12 = "2.1.0-beta.12"
VERSION_STRING_2_1_0_beta_14 = "2.1.0-beta.14"

# leading zeroes in the beta number can break things
VERSION_STRING_2_18_0_beta_01 = "2.18.0-beta.01"


def test_cli_version_01():
    v200 = OPCLIVersion(VERSION_STRING_2_0_0)
    v210 = OPCLIVersion(VERSION_STRING_2_1_0)
    assert v200 < v210


def test_cli_version_02():
    v200 = OPCLIVersion(VERSION_STRING_2_0_0)

    assert v200.is_beta is False


def test_cli_version_03():
    v200 = OPCLIVersion(VERSION_STRING_2_0_0)
    v200_beta12 = OPCLIVersion(VERSION_STRING_2_0_0_beta_01)
    assert v200 > v200_beta12


def test_cli_version_04():
    v200_beta12 = OPCLIVersion(VERSION_STRING_2_0_0_beta_01)
    assert v200_beta12 < VERSION_STRING_2_0_0


def test_cli_version_05():
    v200 = OPCLIVersion(VERSION_STRING_2_0_0)
    assert v200 == VERSION_STRING_2_0_0


def test_cli_version_06():
    v200_beta12 = OPCLIVersion(VERSION_STRING_2_0_0_beta_01)
    v200_beta12_str = str(v200_beta12)
    print(VERSION_STRING_2_0_0_beta_01)
    print(v200_beta12_str)
    assert v200_beta12_str == VERSION_STRING_2_0_0_beta_01


def test_cli_version_07():
    v200_beta12 = OPCLIVersion(VERSION_STRING_2_0_0_beta_01)
    v210_beta12 = OPCLIVersion(VERSION_STRING_2_1_0_beta_12)
    assert v200_beta12 < v210_beta12


def test_cli_version_08():
    v200_a = OPCLIVersion(VERSION_STRING_2_0_0)
    v200_b = v200_a

    assert not v200_a < v200_b


def test_cli_version_09():
    v200 = OPCLIVersion(VERSION_STRING_2_0_0)
    v210 = OPCLIVersion(VERSION_STRING_2_1_0)
    assert v200 <= v210


def test_cli_version_10():
    v200_beta12 = OPCLIVersion(VERSION_STRING_2_0_0_beta_01)
    v200_beta14 = OPCLIVersion(VERSION_STRING_2_0_0_beta_14)
    assert v200_beta14 > v200_beta12


def test_cli_version_11():
    v200_beta12 = OPCLIVersion(VERSION_STRING_2_0_0_beta_01)
    v200_beta14 = OPCLIVersion(VERSION_STRING_2_0_0_beta_14)
    assert v200_beta14 > v200_beta12


def test_cli_version_12():
    v200_beta12 = OPCLIVersion(VERSION_STRING_2_0_0_beta_01)
    v210 = OPCLIVersion(VERSION_STRING_2_1_0)
    assert v200_beta12 < v210


def test_cli_version_13():
    v210 = OPCLIVersion(VERSION_STRING_2_1_0)
    v210_beta12 = OPCLIVersion(VERSION_STRING_2_1_0_beta_12)
    assert v210_beta12 < v210


def test_cli_version_14():
    v2_18_0_beta_01 = OPCLIVersion(VERSION_STRING_2_18_0_beta_01)
    v2_18_0_beta_01_str = str(v2_18_0_beta_01)
    print(VERSION_STRING_2_18_0_beta_01)
    print(v2_18_0_beta_01_str)
    assert v2_18_0_beta_01_str == VERSION_STRING_2_18_0_beta_01



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from pyonepassword.api.exceptions import (
    OPFieldNotFoundException,
    OPInvalidItemException,
    OPItemFieldCollisionException,
    OPSectionCollisionException,
    OPUnknownItemTypeException
)
from pyonepassword.api.object_types import OPLoginItem
from pyonepassword.op_items import OPItemFactory

if TYPE_CHECKING:
    from .fixtures.invalid_data import InvalidData
    from .fixtures.valid_data import ValidData


# ensure HOME env variable is set, and there's a valid op config present
pytestmark = pytest.mark.usefixtures("valid_op_cli_config_homedir")


def test_unknown_item_type_01(invalid_data):
    invalid_item_json = invalid_data.data_for_name("invalid-item")
    with pytest.raises(OPUnknownItemTypeException):
        _ = OPItemFactory.op_item(invalid_item_json)


def test_malformed_item_json_01(invalid_data):
    malformed_json = invalid_data.data_for_name("malformed-item-json")
    with pytest.raises(OPInvalidItemException):
        _ = OPItemFactory.op_item(malformed_json)


def test_item_field_not_found_01(valid_data: ValidData):
    item_dict = valid_data.data_for_name("example-login-1")
    result = OPLoginItem(item_dict)

    with pytest.raises(OPFieldNotFoundException):
        result.field_by_id("Non-existent-field")


def test_item_field_collision_01(invalid_data):
    field_collision_json = invalid_data.data_for_name("field-collision")
    with pytest.raises(OPItemFieldCollisionException):
        OPItemFactory.op_item(field_collision_json)


def test_item_section_collision_01(invalid_data: InvalidData):
    """
    Test item creation with colliding sections

    Create:
        - op item object from an item dictionary with colliding sections
    Verify:
        - OPSectionCollisionException is raised
    """
    invalid_item_dict = invalid_data.data_for_name(
        "login-item-with-section-collision")
    with pytest.raises(OPSectionCollisionException):
        OPItemFactory.op_item(invalid_item_dict)



from typing import Optional

from .._item_descriptor_registry import op_register_item_descriptor_type
from .._item_type_registry import op_register_item_type
from ._item_base import OPAbstractItem
from ._item_descriptor_base import OPAbstractItemDescriptor


@op_register_item_descriptor_type
class OPSSHKeyItemDescriptor(OPAbstractItemDescriptor):
    CATEGORY = "SSH_KEY"

    def __init__(self, item_dict):
        super().__init__(item_dict)

    @property
    def fingerprint(self) -> str:
        return self["additional_information"]


@op_register_item_type
class OPSSHKeyItem(OPAbstractItem):
    CATEGORY = "SSH_KEY"

    def __init__(self, item_dict):
        super().__init__(item_dict)

    @property
    def public_key(self) -> str:
        password = self.field_value_by_id("public_key")
        return password

    @property
    def public_key_reference(self) -> Optional[str]:
        password = self.field_reference_by_id("public_key")
        return password

    @property
    def fingerprint(self) -> str:
        return self.field_value_by_id("fingerprint")

    @property
    def fingerprint_reference(self) -> Optional[str]:
        return self.field_reference_by_id("fingerprint")

    @property
    def private_key(self) -> str:
        return self.field_value_by_id("private_key")

    @property
    def private_key_reference(self) -> Optional[str]:
        return self.field_reference_by_id("private_key")

    @property
    def key_type(self) -> str:
        return self.field_value_by_id("key_type")

    @property
    def key_type_reference(self) -> Optional[str]:
        return self.field_reference_by_id("key_type")


@op_register_item_type
class OPSSHKeyItemRelaxedValidation(OPSSHKeyItem):
    _relaxed_validation = True



from __future__ import annotations

from typing import TYPE_CHECKING, List

import pytest

from pyonepassword.api.object_types import (
    OPConcealedField,
    OPLoginItem,
    OPStringField,
    OPTOTPField
)
from pyonepassword.op_items.field_registry import OPItemFieldFactory

if TYPE_CHECKING:
    from .fixtures.expected_item_fields import (
        ExpectedItemField,
        ExpectedItemFieldData
    )
    from .fixtures.expected_login import ExpectedLogin
    from .fixtures.valid_data import ValidData


def _lookup_exepcted_item_field(fields: List[ExpectedItemField], field_id: str):
    """
    Field order may not be reliable, so we have to hunt for the right one
    """
    field = None
    for f in fields:
        if f.field_id == field_id:
            field = f
    return field


# ensure HOME env variable is set, and there's a valid op config present
pytestmark = pytest.mark.usefixtures("valid_op_cli_config_homedir")


def test_item_field_01(valid_data: ValidData, expected_item_field_data: ExpectedItemFieldData):
    expected: ExpectedItemField

    expected = expected_item_field_data.data_for_key("example-login-username")
    field_dict = valid_data.data_for_name("login-item-field-username")
    field = OPItemFieldFactory.item_field(field_dict)
    assert isinstance(field, OPStringField)
    assert field.field_id == expected.field_id


def test_item_field_02(valid_data: ValidData, expected_item_field_data: ExpectedItemFieldData):
    expected: ExpectedItemField

    expected = expected_item_field_data.data_for_key("example-login-username")
    field_dict = valid_data.data_for_name("login-item-field-username")
    field = OPItemFieldFactory.item_field(field_dict)
    assert field.field_type == expected.type


def test_item_field_03(valid_data: ValidData, expected_item_field_data: ExpectedItemFieldData):
    expected: ExpectedItemField

    expected = expected_item_field_data.data_for_key("example-login-username")
    field_dict = valid_data.data_for_name("login-item-field-username")
    field = OPItemFieldFactory.item_field(field_dict)
    assert field.reference is not None
    assert field.reference == expected.reference


def test_item_field_04(valid_data: ValidData, expected_item_field_data: ExpectedItemFieldData):
    expected: ExpectedItemField

    expected = expected_item_field_data.data_for_key("example-login-username")
    field_dict = valid_data.data_for_name("login-item-field-username")
    field = OPItemFieldFactory.item_field(field_dict)
    assert field.label == expected.label


def test_item_field_05(valid_data: ValidData, expected_item_field_data: ExpectedItemFieldData):
    expected: ExpectedItemField

    expected = expected_item_field_data.data_for_key("example-login-username")
    field_dict = valid_data.data_for_name("login-item-field-username")
    field = OPItemFieldFactory.item_field(field_dict)
    assert field.value == expected.value


def test_item_field_06(valid_data: ValidData, expected_item_field_data: ExpectedItemFieldData):
    expected: ExpectedItemField

    expected = expected_item_field_data.data_for_key("example-login-username")
    field_dict = valid_data.data_for_name("login-item-field-username")
    field = OPItemFieldFactory.item_field(field_dict)
    assert field.purpose == expected.purpose


def test_item_field_11(valid_data: ValidData, expected_item_field_data: ExpectedItemFieldData):
    expected: ExpectedItemField

    expected = expected_item_field_data.data_for_key("example-login-password")
    field_dict = valid_data.data_for_name("login-item-field-password")
    field = OPItemFieldFactory.item_field(field_dict)
    assert isinstance(field, OPConcealedField)
    assert field.field_id == expected.field_id


def test_item_field_12(valid_data: ValidData, expected_item_field_data: ExpectedItemFieldData):
    expected: ExpectedItemField

    expected = expected_item_field_data.data_for_key("example-login-password")
    field_dict = valid_data.data_for_name("login-item-field-password")
    field = OPItemFieldFactory.item_field(field_dict)
    assert field.field_type == expected.type


def test_item_field_13(valid_data: ValidData, expected_item_field_data: ExpectedItemFieldData):
    expected: ExpectedItemField

    expected = expected_item_field_data.data_for_key("example-login-password")
    field_dict = valid_data.data_for_name("login-item-field-password")
    field = OPItemFieldFactory.item_field(field_dict)
    assert field.reference is not None
    assert field.reference == expected.reference


def test_item_field_14(valid_data: ValidData, expected_item_field_data: ExpectedItemFieldData):
    expected: ExpectedItemField

    expected = expected_item_field_data.data_for_key("example-login-password")
    field_dict = valid_data.data_for_name("login-item-field-password")
    field = OPItemFieldFactory.item_field(field_dict)
    assert field.label == expected.label


def test_item_field_15(valid_data: ValidData, expected_item_field_data: ExpectedItemFieldData):
    expected: ExpectedItemField

    expected = expected_item_field_data.data_for_key("example-login-password")
    field_dict = valid_data.data_for_name("login-item-field-password")
    field = OPItemFieldFactory.item_field(field_dict)
    assert field.value == expected.value


def test_item_field_16(valid_data: ValidData, expected_item_field_data: ExpectedItemFieldData):
    expected: ExpectedItemField

    expected = expected_item_field_data.data_for_key("example-login-password")
    field_dict = valid_data.data_for_name("login-item-field-password")
    field = OPItemFieldFactory.item_field(field_dict)
    assert field.purpose == expected.purpose


def test_item_field_17(valid_data: ValidData, expected_item_field_data: ExpectedItemFieldData):
    expected: ExpectedItemField
    expected = expected_item_field_data.data_for_key("example-login-password")
    field_dict = valid_data.data_for_name("login-item-field-password")
    field = OPItemFieldFactory.item_field(field_dict)
    assert isinstance(field.entropy, float)
    margin = .000001
    expected_entropy = expected.entropy
    high = expected_entropy * (1 + margin)
    low = expected_entropy * (1 - margin)
    assert field.entropy < high
    assert field.entropy > low


def test_item_field_18(valid_data: ValidData, expected_item_field_data: ExpectedItemFieldData):
    expected: ExpectedItemField

    expected = expected_item_field_data.data_for_key(
        "example-totp-field-no-issuer")
    field_dict = valid_data.data_for_name("login-item-field-totp")
    field: OPTOTPField = OPItemFieldFactory.item_field(field_dict)
    assert isinstance(field, OPTOTPField)
    assert field.totp == expected.totp


def test_item_field_19(valid_data: ValidData, expected_item_field_data: ExpectedItemFieldData):
    expected: ExpectedItemField

    expected = expected_item_field_data.data_for_key(
        "example-totp-field-no-issuer")
    field_dict = valid_data.data_for_name("login-item-field-totp")
    field: OPTOTPField = OPItemFieldFactory.item_field(field_dict)
    assert isinstance(field, OPTOTPField)
    assert field.totp_secret == expected.value


def test_item_lookup_field_01(valid_data: ValidData, expected_login_item_data):
    item_name = "Example Login with Fields"
    field_label = "Example Field"
    field_id = "zyhpoqkrwjtrq6qh5vhwylh6ie"
    expected_login: ExpectedLogin = expected_login_item_data.data_for_login(
        item_name)

    expected_field_list: ExpectedItemField = expected_login.fields_by_label(
        field_label)
    expected_field = _lookup_exepcted_item_field(expected_field_list, field_id)

    valid_item_dict = valid_data.data_for_name("example-login-with-fields")
    result_login_item = OPLoginItem(valid_item_dict)
    result = result_login_item.fields_by_label(field_label)
    result = result[1]
    assert result.value == expected_field.value


def test_item_lookup_field_02(valid_data: ValidData, expected_login_item_data):
    item_name = "Example Login with Fields"
    field_label = "Example Field"
    field_id = "ikr76mnggw767qwqoel624oqv4"
    expected_login: ExpectedLogin = expected_login_item_data.data_for_login(
        item_name)

    expected_field_list: ExpectedItemField = expected_login.fields_by_label(
        field_label)
    expected_field = _lookup_exepcted_item_field(expected_field_list, field_id)

    valid_item_dict = valid_data.data_for_name("example-login-with-fields")
    result_login_item = OPLoginItem(valid_item_dict)
    result = result_login_item.first_field_by_label(field_label)

    assert result.value == expected_field.value


def test_item_lookup_field_03(valid_data: ValidData, expected_login_item_data):
    """
    Test case-insensitive field lookup by label

    Create:
        - a login item with fields and sections
        - case-insensitively look up a field via first_field_by_label()
    Verify:
        - the lookup succeeded
        - the resulting field matches the expected value
    """
    item_name = "Example Login with Fields"
    field_label = "Example Field"
    field_id = "ikr76mnggw767qwqoel624oqv4"
    expected_login: ExpectedLogin = expected_login_item_data.data_for_login(
        item_name)

    expected_field_list: ExpectedItemField = expected_login.fields_by_label(
        field_label)
    expected_field = _lookup_exepcted_item_field(expected_field_list, field_id)

    valid_item_dict = valid_data.data_for_name("example-login-with-fields")
    result_login_item = OPLoginItem(valid_item_dict)
    field_label_lower = field_label.lower()
    result = result_login_item.first_field_by_label(
        field_label_lower, case_sensitive=False)

    assert result.value == expected_field.value



import pytest

from pyonepassword._datetime import fromisoformat_z


def test_datetime_isoformat_no_z_01():
    invalid_date_str = "2021-06-29T18:31:30"
    with pytest.raises(ValueError):
        fromisoformat_z(invalid_date_str)


def test_datetime_isoformat_utc_zterm_01(expected_datetime_data):
    datetime_str = "2021-06-29T18:31:30Z"
    expected_datetime = expected_datetime_data.datetime_for_key(
        "example-datetime")
    result = fromisoformat_z(datetime_str)
    assert result == expected_datetime


def test_datetime_isoformat_utc_zeros_01(expected_datetime_data):
    datetime_str = "2021-06-29T18:31:30+00:00"
    expected_datetime = expected_datetime_data.datetime_for_key(
        "example-datetime")
    result = fromisoformat_z(datetime_str)
    assert result == expected_datetime



from ..json import safe_unjson


class OPTOTPItem(dict):
    """
    {
        "id": "TOTP_obe3whgcenxulnxs5cjyjy763i",
        "section": {
            "id": "add more"
        },
        "type": "OTP",
        "label": "one-time password",
        "value": "AAAAY3DPEHPK3PXP",
        "totp": "368006",
        "reference": "op://Test Data/Login With TOTP/add more/one-time password"
    }
    """

    def __init__(self, totp_dict_or_json):
        totp_dict = safe_unjson(totp_dict_or_json)
        super().__init__(totp_dict)

    @property
    def unique_id(self) -> str:
        return self["id"]

    @property
    def totp(self) -> str:
        return self["totp"]

    @property
    def value(self) -> str:
        return self["value"]

    @property
    def label(self) -> str:
        return self["label"]

    @property
    def reference(self) -> str:
        return self["reference"]



from __future__ import annotations

from json import JSONDecodeError
from typing import TYPE_CHECKING, Any, Dict, Optional, Type, Union

from ..json import safe_unjson
from ..py_op_exceptions import OPInvalidFieldException
from .fields_sections.item_field_base import OPItemField
from .uuid import OPUniqueIdentifierBase32, is_uuid

if TYPE_CHECKING:
    from .fields_sections.item_section import OPSection  # pragma: no coverage


class OPUnknownFieldTypeException(Exception):
    def __init__(self, msg, item_dict=None):
        super().__init__(msg)
        self.item_dict = item_dict


class OPNewItemField(OPItemField):
    """
    A class for creating new fields for use with new items

    New fields may be created directly, or from existing fields
    """
    FIELD_TYPE: Optional[str] = None
    FIELD_PURPOSE: Optional[str] = None

    def __init__(self, field_label: str, value: Any, field_id: str = None, section: Optional[OPSection] = None):
        """
        Create a new field object

        NOTE: This class must be subclassed and FIELD_TYPE overridden

        Parameters
        ----------
        field_label: str
            The user-visible name of the field
        field_id: str, optional
            The unique identifier for this field. If none is provided, a random one will be generated
        section: OPSection, optional
            The section this field should be associated with. Not all fields are in sections
        """
        if not self.FIELD_TYPE:  # pragma: no cover
            raise TypeError(
                f"{self.__class__.__name__} must be overridden and FIELD_TYPE set")

        if not field_id:
            unique_id = OPUniqueIdentifierBase32()
            field_id = str(unique_id)
        field_dict = {
            "id": field_id,
            "label": field_label,
            "value": value,
            "type": self.FIELD_TYPE
        }
        if self.FIELD_PURPOSE:
            field_dict["purpose"] = self.FIELD_PURPOSE
        if section:
            field_dict["section"] = dict(section)
        super().__init__(field_dict)
        if section:
            section.register_field(self)

    def update_section(self, section: OPSection):
        """
        Update a field's associated section in the event
        a section's UUID was regenerated
        """
        if self.section_id != section.section_id:
            self["section"] = dict(section)
            section.register_field(self)

    @classmethod
    def from_field(cls, field: OPItemField, section: Optional[OPSection] = None):
        """
        Create a new field from an existing one

        If the existing field's ID is a random ID, a new random ID will be generated

        Parameters
        ----------
        field: OPItemField
            The field object to duplicate
        section: OPSection, optional
            The section this field should be associated with. Not all fields are in sections
        Returns
        -------
        new_field: OPItemField
            The newly created field object
        """
        field_id = field["id"]
        if is_uuid(field_id):
            field_id = str(OPUniqueIdentifierBase32())
        label = field["label"]
        value = field["value"]
        new_field = cls(label, value, field_id=field_id, section=section)
        return new_field


class OPNewItemFieldFactory:
    _TYPE_REGISTRY: Dict[str, Type[OPNewItemField]] = {}

    @classmethod
    def register_op_field_type(cls, item_class):
        cls._TYPE_REGISTRY[item_class.FIELD_TYPE] = item_class

    @classmethod
    def _field_from_dict(cls, field_dict, section):
        item_type = field_dict["type"]

        try:
            item_cls = cls._TYPE_REGISTRY[item_type]
        except KeyError as ke:
            raise OPUnknownFieldTypeException(
                f"Unknown item type {item_type}", item_dict=field_dict) from ke

        return item_cls.from_field(field_dict, section=section)

    @classmethod
    def item_field(cls, item_json_or_dict: Union[str, Dict], section: Optional[OPSection] = None):
        try:
            field_dict = safe_unjson(item_json_or_dict)
        except JSONDecodeError as jdce:
            raise OPInvalidFieldException(
                f"Failed to unserialize field JSON: {jdce}") from jdce
        obj = cls._field_from_dict(field_dict, section)
        return obj


def op_register_new_item_field_type(item_class):
    OPNewItemFieldFactory.register_op_field_type(item_class)
    return item_class



from datetime import datetime


def fromisoformat_z(date_string: str) -> datetime:
    """
    Python's datetime.datetime.fromisoformat() can't parse datetime strings
    With a UTC 'Z' timezone suffix instead of +00:00
    And there's no way to do it without an external package like dateutil
    https://bugs.python.org/issue35829
    """

    if date_string.endswith('Z'):
        _str = date_string.rstrip('Z')
        _str += "+00:00"
    elif date_string.endswith("+00:00"):
        # We shouldn't ever get this, but we also shouldn't blow up on it,
        # so work with it if it happens
        _str = date_string
    else:
        # Mirror the ValueError that datetime.fromisoformat() raises
        raise ValueError(
            f"Invalid Z-terminated isoformat string: '{date_string}'")
    datetime_obj = datetime.fromisoformat(_str)
    return datetime_obj



from __future__ import annotations

from typing import TYPE_CHECKING, List

import pytest

from pyonepassword.api.exceptions import OPSectionNotFoundException
from pyonepassword.api.object_types import OPLoginItem
from pyonepassword.op_items.fields_sections.item_field_base import OPItemField
from pyonepassword.op_items.fields_sections.item_section import OPSection

if TYPE_CHECKING:
    from .fixtures.expected_item_fields import ExpectedItemField
    from .fixtures.expected_login import ExpectedLogin, ExpectedLoginItemData
    from .fixtures.valid_data import ValidData


# ensure HOME env variable is set, and there's a valid op config present
pytestmark = pytest.mark.usefixtures("valid_op_cli_config_homedir")


def _lookup_exepcted_item_field(fields: List[ExpectedItemField], field_id: str) -> ExpectedItemField:
    """
    Field order may not be reliable, so we have to hunt for the right one
    """
    field = None
    for f in fields:
        if f.field_id == field_id:
            field = f
    return field


def test_item_section_01(valid_data: ValidData, expected_login_item_data: ExpectedLoginItemData):
    """
    Test search for first field by label

    Create:
      - Look up a section on the item object by its ID
      - Look up a field on the seciton object using first_field_by_label()
    Verify:
      - the lookup doesn't fail
      - the returned section's value matches the expected value
    """
    item_name = "Example Login with Fields"
    section_id = "vh4wk7qyw46urc7wuwczzhpm7u"
    field_label = "Example Field"
    field_id = "ikr76mnggw767qwqoel624oqv4"
    expected_login: ExpectedLogin = expected_login_item_data.data_for_login(
        item_name)

    expected_field_list: ExpectedItemField = expected_login.fields_by_label(
        field_label)
    expected_field = _lookup_exepcted_item_field(expected_field_list, field_id)
    valid_item_dict = valid_data.data_for_name("example-login-with-fields")
    result_login_item = OPLoginItem(valid_item_dict)

    result_item_section = result_login_item.section_by_id(section_id)

    result = result_item_section.first_field_by_label(field_label)

    assert isinstance(result, OPItemField)
    assert result.value == expected_field.value


def test_item_section_02(valid_data: ValidData, expected_login_item_data: ExpectedLoginItemData):
    """
    Test case-insensitive search for first field by label

    Create:
      - A login item object with fields and sections
      - Look up a section on the item object by its ID
      - Case-insensitive look up a field on the section object using first_field_by_label()
    Verify:
      - the lookup doesn't fail
      - the returned field

    """
    item_name = "Example Login with Fields"
    section_id = "vh4wk7qyw46urc7wuwczzhpm7u"
    field_label = "Example Field"
    field_id = "ikr76mnggw767qwqoel624oqv4"
    expected_login: ExpectedLogin = expected_login_item_data.data_for_login(
        item_name)

    expected_field_list: ExpectedItemField = expected_login.fields_by_label(
        field_label)
    expected_field = _lookup_exepcted_item_field(expected_field_list, field_id)
    valid_item_dict = valid_data.data_for_name("example-login-with-fields")
    result_login_item = OPLoginItem(valid_item_dict)

    result_item_section = result_login_item.section_by_id(section_id)

    result = result_item_section.first_field_by_label(
        field_label, case_sensitive=False)

    assert isinstance(result, OPItemField)
    assert result.value == expected_field.value


def test_item_section_03(valid_data: ValidData, expected_login_item_data: ExpectedLoginItemData):
    """
    Test looking up a section on a login item by an invalid section ID

    Create:
        - A login item object with fields and sections
        - Look up an invalid section ID via section_by_id()
    Verify:
        - OPSectionNotFoundException is raised
    """
    section_id = "no_such_section"

    valid_item_dict = valid_data.data_for_name("example-login-with-fields")
    result_login_item = OPLoginItem(valid_item_dict)

    with pytest.raises(OPSectionNotFoundException):
        result_login_item.section_by_id(section_id)


def test_item_section_04(valid_data: ValidData, expected_login_item_data: ExpectedLoginItemData):
    """
    Test case-insensitive search for sections by label

    Create:
      - A login item object with fields and sections
      - Look up a section on the item object by its ID
      - Case-insensitive look up a field on the section object using sections_by_label()
    Verify:
      - the lookup doesn't fail
      - the returned field

    """
    item_name = "Example Login with Fields"
    section_id = "vh4wk7qyw46urc7wuwczzhpm7u"
    section_label = "Example Section 1"

    expected_login: ExpectedLogin = expected_login_item_data.data_for_login(
        item_name)

    expected_section = expected_login.section_by_id(section_id)

    valid_item_dict = valid_data.data_for_name("example-login-with-fields")
    result_login_item = OPLoginItem(valid_item_dict)

    section_label_lower = section_label.lower()
    result = result_login_item.sections_by_label(
        section_label_lower, case_sensitive=False)[0]

    assert isinstance(result, OPSection)
    assert result.label == expected_section.label



try:
    # register colored_traceback hook if in a dev environment (dev-reqs.txt)
    import colored_traceback.auto
except ImportError:  # pragma: no coverage
    # otherwise just stick to the basics
    pass

from .__about__ import __summary__, __title__, __version__
# Import and discard _API_INITIALIZER to ensure all API classes get registered
from ._api_initializer import _API_INITIALIZED as _
# to re-export, either use this redundant import-as construct
# or do:
# __all__ = ["OP"]
# https://mypy.readthedocs.io/en/stable/config_file.html?highlight=export#confval-implicit_reexport
from .pyonepassword import OP as OP



import fnmatch
import logging
from os import environ as env
from typing import List, Optional, Type, Union

from ._py_op_commands import (
    EXISTING_AUTH_IGNORE,
    ExistingAuthEnum,
    _OPCommandInterface
)
from ._py_op_deprecation import deprecated_kwargs
from .account import OPAccountList
from .op_items._item_list import OPItemList
from .op_items._item_type_registry import OPItemFactory
from .op_items._new_item import OPNewItemMixin
from .op_items.item_types._item_base import OPAbstractItem
from .op_items.item_types.generic_item import (
    _OPGenericItem,
    _OPGenericItemRelaxedValidation
)
from .op_items.item_types.login import (
    OPLoginItemNewPrimaryURL,
    OPLoginItemTemplate
)
from .op_items.password_recipe import OPPasswordRecipe
from .op_items.totp import OPTOTPItem
from .op_objects import (
    OPGroup,
    OPGroupDescriptorList,
    OPUser,
    OPUserDescriptorList,
    OPVault,
    OPVaultDescriptorList
)
from .py_op_exceptions import (
    OPCmdFailedException,
    OPDocumentDeleteException,
    OPDocumentGetException,
    OPForgetException,
    OPInvalidDocumentException,
    OPInvalidItemException,
    OPItemDeleteException,
    OPItemDeleteMultipleException,
    OPItemGetException,
    OPItemListException,
    OPSignoutException
)
from .version import PyOPAboutMixin


class OP(_OPCommandInterface, PyOPAboutMixin):
    """
    Class for logging into and querying a 1Password account via the 'op' cli command.
    """
    @deprecated_kwargs(use_existing_session='existing_auth',
                       account_shorthand='account')
    def __init__(self,
                 account: Optional[str] = None,
                 account_shorthand: Optional[str] = None,
                 password: Optional[str] = None,
                 existing_auth: ExistingAuthEnum = EXISTING_AUTH_IGNORE,
                 use_existing_session: bool = False,
                 password_prompt: bool = True,
                 vault: Optional[str] = None,
                 op_path: str = 'op',
                 logger: Optional[logging.Logger] = None):
        """
        Create an OP object. The 1Password (non-initial) sign-in happens during object instantiation.

        Parameters
        ----------
        account : str, optional
            May be any account identifier accepted by the op --account flag:
            account shorthand, sign-in address, account ID, or user ID
        password : str, optional
            If provided, the password will be piped to the 'op' command over stdin
        existing_auth : ExistingAuthEnum
            Whether existing authentication must be used, used if possible, or ignored
            Valid values:
              - EXISTING_AUTH_AVAIL: Use existing authentication if available
              - EXISTING_AUTH_IGNORE: Reauthenticate, ignoring any existing authentication
              - EXISTING_AUTH_REQD: Use existing authentication, failing if there isn't a valid session available
        password_prompt : bool
            Whether an interactive password prompt on the console should be presented if necessary
        vault : str, optional
            If set, this becomes the default argument to the --vault flag
            for future queries.
        op_path : str, optional
            Optional path to the `op` command, if it's not at the default location
        logger : logging.Logger
            A logging object. If not provided a basic logger is created and used

        Raises
        ------
        OPSigninException
            If 1Password sign-in fails for any reason.
        OPAuthenticationException
            if:
                - No previous authentication (e.g., existing session or service account otken) is available
                - no password provided,
                - interactive password prompt is supressed, and
                - biometric authenticaiotn is disabled
        OPNotFoundException
            If the 1Password command can't be found
        """
        super().__init__(vault=vault,
                         account=account,
                         password=password,
                         logger=logger,
                         op_path=op_path,
                         existing_auth=existing_auth,
                         password_prompt=password_prompt)

    def item_get(self, item_identifier, vault=None, include_archive=False, generic_okay=False, relaxed_validation=False) -> OPAbstractItem:
        """
        Get an 'item' object from a 1Password vault.
        The returned object may be any of the item types extending OPAbstractItem.
        These currently include:
        - OPLoginItem
        - OPCreditCardItem
        - OPSecureNoteItem
        - OPPasswordItem
        - OPDocumentItem
        - OPServerItem
        - OPDatabaseItem

        Note that getting a document item is not the same as getting the document itself. The
        item only contains metadata about the document such as filename.

        Parameters
        ----------
        item_identifier: str
            Name or ID of the item to look up
        vault: str, optional
            The name or ID of a vault to override the object's default vault, by default None
        include_archive: bool, optional
            Include items in the Archive, by default False
        generic_okay: bool, optional
            Instantiate unknown item types as _OPGenericItem rather than raise OPUnknownItemException
        relaxed_validation: bool, optional
            Whether to enable relaxed item validation for this query, in order to parse non-conformant data
            by default False
        Note:
            If a non-unique item identifier is provided (e.g., item name/title), and there
            is more than one item that matches, OPItemGetException will be raised. Check the
            error message in OPItemGetException.err_output for details

        Raises
        ------
        OPItemGetException
            If the lookup fails for any reason during command execution
        OPInvalidItemException
            If the item JSON fails to decode
        OPUnknownItemTypeException
            If the item object returned by 1Password isn't a known type and generic_okay is False
        OPNotFoundException
            If the 1Password command can't be found
        Returns
        -------
        item: OPAbstractItem
            An item object of one of the types listed above

        Service Account Support
        -----------------------
        Supported
          required keyword arguments: vault
        """

        output = super()._item_get(item_identifier, vault=vault,
                                   decode="utf-8", include_archive=include_archive)
        op_item = OPItemFactory.op_item(
            output, generic_okay=generic_okay, relaxed_validation=relaxed_validation)
        return op_item

    def item_get_totp(self, item_identifier: str, vault=None) -> OPTOTPItem:
        """
        Get a TOTP code from the item specified by name or UUID.

        Note: Items in the Archive are ignored by default. To get the TOTP for an
        item in the Archive, specify the item by UUID.

        Parameters
        ----------
        item_identifier: str
            Name or ID of the item to look up
        vault: str, optional
            The name or ID of a vault to override the object's default vault

        Note:
            If a non-unique item identifier is provided (e.g., item name/title), and there
            is more than one item that matches, OPItemGetException will be raised. Check the
            error message in OPItemGetException.err_output for details

        Raises
        ------
        OPItemGetException
            If the lookup fails for any reason during command execution
        OPNotFoundException
            If the 1Password command can't be found

        Returns
        -------
        totp_code: str
            A string representing the TOTP code

        Service Account Support
        -----------------------
        Supported
          required keyword arguments: vault
        """
        output = super()._item_get_totp(item_identifier, vault=vault, decode="utf-8")
        # strip newline
        totp = OPTOTPItem(output)
        return totp

    def user_get(self, user_name_or_id: str) -> OPUser:
        """
        Return the details for the user specified by name or UUID.

        Parameters
        ----------
        user_name_or_id: str
            Name or ID of the user to look up
        Raises
        ------
        OPUserGetException
            If the lookup fails for any reason during command execution
        OPNotFoundException
            If the 1Password command can't be found

        Returns
        -------
        user: OPuser
            An object representing the details of the requested user

        Service Account Support
        -----------------------
        Supported
        """
        user_json = super()._user_get(user_name_or_id)
        user = OPUser(user_json)
        return user

    def user_list(self, group_name_or_id=None, vault_name_or_id=None) -> OPUserDescriptorList:
        """
        Return a list of users in an account.

        Parameters
        ----------
        group_name_or_id: str
            Name or ID of a group to restrict user listing to
        vault_name_or_id: str
            Name or ID of a vault to restrict user listing to

        Raises
        ------
        OPUserListException
            If the user list operation for any reason during command execution
        OPNotFoundException
            If the 1Password command can't be found

        Returns
        -------
        user: OPUserDescriptorList
            An object representing a list of user descriptors

        Service Account Support
        -----------------------
        Supported
        """
        user_list: Union[str, OPUserDescriptorList]

        user_list = self._user_list(
            group_name_or_id=group_name_or_id, vault=vault_name_or_id)
        user_list = OPUserDescriptorList(user_list)
        return user_list

    def vault_get(self, vault_name_or_id: str) -> OPVault:
        """
        Return the details for the vault specified by name or UUID.

        Parameters
        ----------
        vault_name_or_id: str
            Name or UUID of the vault to look up

        Raises
        ------
        OPVaultGetException
            If the lookup fails for any reason during command execution
        OPNotFoundException
            If the 1Password command can't be found

        Returns
        -------
        vault: OPVault
            An object representing the details of the requested vault

        Service Account Support
        -----------------------
        Supported
            prohibited keyword arguments: group, user
        """
        vault_json = super()._vault_get(vault_name_or_id, decode="utf-8")
        vault = OPVault(vault_json)
        return vault

    def vault_list(self, group_name_or_id=None, user_name_or_id=None) -> OPVaultDescriptorList:
        """
        Return a list of vaults in an account.

        Parameters
        ----------
        group_name_or_id: str
            Name or ID of a group to restrict vault listing to
        user_name_or_id: str
            Name or ID of a user to restrict vault listing to

        Raises
        ------
        OPVaultListException
            If the vault list operation for any reason during command execution
        OPNotFoundException
            If the 1Password command can't be found

        Returns
        -------
        user: OPVaultDescriptorList
            An object representing a list of vault descriptors

        Service Account Support
        -----------------------
        Supported
        """
        vault_list_json = super()._vault_list(
            group_name_or_id=group_name_or_id, user_name_or_id=user_name_or_id)
        vault_list = OPVaultDescriptorList(vault_list_json)
        return vault_list

    def group_get(self, group_name_or_id: str) -> OPGroup:
        """
        Return the details for the group specified by name or UUID.

        Parameters
        ----------
        group_name_or_id: str
            Name or UUID of the group to look up

        Raises
        ------
        OPGroupGetException
            If the lookup fails for any reason during command execution
        OPNotFoundException
            If the 1Password command can't be found

        Returns
        -------
        user: OPGroup
            An object representing the details of the requested group

        Service Account Support
        -----------------------
        Supported
        """
        group_json = super()._group_get(group_name_or_id, decode="utf-8")
        group = OPGroup(group_json)
        return group

    def group_list(self, user_name_or_id=None, vault=None) -> OPGroupDescriptorList:
        """
        Return a list of groups in an account.

        Parameters
        ----------
        user_name_or_id: str
            Name or ID of a user to restrict vault listing to
        group_name_or_id: str
            Name or ID of a group to restrict vault listing to

        Raises
        ------
        OPGroupListException
            If the group list operation for any reason during command execution
        OPNotFoundException
            If the 1Password command can't be found

        Returns
        -------
        user: OPGroupDescriptorList
            An object representing a list of vault descriptors

        Service Account Support
        -----------------------
        Supported
        """
        group_list: Union[str, OPGroupDescriptorList]
        group_list = self._group_list(
            user_name_or_id=user_name_or_id, vault=vault)
        group_list = OPGroupDescriptorList(group_list)
        return group_list

    def item_get_password(self, item_identifier, vault=None, relaxed_validation=False) -> str:
        """
        Get the value of the password field from the item specified by name or UUID.

        Parameters
        ----------
        item_identifier: str
            The item to look up
        vault: str, optional
            The name or ID of a vault to override the object's default vault
        relaxed_validation: bool, optional
            Whether to enable relaxed item validation for this query, in order to parse non-conformant data
            by default False

        Raises
        ------
        AttributeError
            If the item doesn't have a 'fileName' attribute.
        OPItemGetException
            If the lookup fails for any reason during command execution
        OPNotFoundException
            If the 1Password command can't be found.

        Returns
        -------
        password: str
            Value of the item's 'password' attribute

        Service Account Support
        -----------------------
        Supported
          required keyword arguments: vault
        """
        item: OPAbstractItem
        item = self.item_get(item_identifier, vault=vault,
                             relaxed_validation=relaxed_validation)

        # satisfy 'mypy': OPAbstractItem has no "password" attribute
        if not hasattr(item, "password"):
            raise OPInvalidItemException(
                f"Item: {item.title} has no password attribute")
        else:
            password = item.password
        return password

    def item_get_filename(self, item_identifier, vault=None, include_archive=False, relaxed_validation=False):
        """
        Get the fileName attribute a document item from a 1Password vault by name or UUID.

        Parameters
        ----------
        item_identifier: str
            The item to look up
        vault: str, optional
            The name or ID of a vault to override the object's default vault
        relaxed_validation: bool, optional
            Whether to enable relaxed item validation for this query, in order to parse non-conformant data
            by default False
        Note:
            If a non-unique item identifier is provided (e.g., item name/title), and there
            is more than one item that matches, OPItemGetException will be raised. Check the
            error message in OPItemGetException.err_output for details

        Raises
        ------
        AttributeError
            If the item doesn't have a 'fileName' attribute
        OPItemGetException
            If the lookup fails for any reason during command execution
        OPNotFoundException
            If the 1Password command can't be found

        Returns
        -------
        file_name: str
            Value of the item's 'fileName' attribute

        Service Account Support
        -----------------------
        Supported
          required keyword arguments: vault
        """
        item = self.item_get(item_identifier, vault=vault,
                             include_archive=include_archive, relaxed_validation=relaxed_validation)

        # raise AttributeError if item isn't a OPDocumentItem
        # we have to raise it ourselves becuase mypy complains OPAbstractItem doesn't have
        # '.file_name'
        if hasattr(item, "file_name"):
            file_name = item.file_name
        else:
            raise AttributeError(
                f"{item.__class__.__name__} object has no attribute 'file_name'")

        return file_name

    def document_get(self, document_name_or_id, vault=None, include_archive=False, relaxed_validation=False):
        """
        Download a document object from a 1Password vault by name or UUID.

        Parameters
        ----------
        document_name_or_id : str
            The item to look up
        vault: str, optional
            The name or ID of a vault to override the object's default vault
        relaxed_validation: bool, optional
            Whether to enable relaxed item validation for this query, in order to parse non-conformant data
            by default False
        Raises
        ------
        OPInvalidDocumentException
            If the retrieved item isn't a document object or lacks a documents expected attributes
        OPDocumentGetException
            If document lookup fails for any reason during command execution
        OPNotFoundException
            If the 1Password command can't be found

        Returns
        -------
        file_name, document bytes: Tuple[str, bytes]
            A tuple consisting of the filename and bytes of the specified document

        Service Account Support
        -----------------------
        Supported
          required keyword arguments: vault
        """
        try:
            file_name = self.item_get_filename(
                document_name_or_id, vault=vault, include_archive=include_archive, relaxed_validation=relaxed_validation)
        except AttributeError as ae:
            raise OPInvalidDocumentException(
                "Item has no 'fileName' attribute") from ae
        except OPCmdFailedException as ocfe:
            raise OPDocumentGetException.from_opexception(ocfe) from ocfe

        try:
            document_bytes = super()._document_get(document_name_or_id,
                                                   vault=vault, include_archive=include_archive)
        except OPCmdFailedException as ocfe:
            raise OPDocumentGetException.from_opexception(ocfe) from ocfe

        return (file_name, document_bytes)

    def document_delete(self, document_identifier: str, vault: Optional[str] = None, archive: bool = False, relaxed_validation: bool = False) -> str:
        """
        Delete a document object based on document name or unique identifier

        Parameters
        ----------
        document_identifier : str
            Name or identifier of the document to delete
        vault : str, optional
            The name or ID of a vault to override the default vault, by default None
        archive : bool, optional
            Whether to archive or permanently delete the item, by default False
        relaxed_validation: bool, optional
            Whether to enable relaxed item validation for this query, in order to parse non-conformant data
            by default False
        Returns
        -------
        document_id: str
            Unique identifier of the item deleted

        Raises
        ------
        OPDocumentDeleteException
            - If the document to be deleted is not found
            - If there is more than one item matching 'document_identifier'
            - If the delete operation fails for any other reason

        Service Account Support
        -----------------------
        Supported
          required keyword arguments: vault
        """

        # to satisfy mypy
        generic_item_class: Type[_OPGenericItem]
        if relaxed_validation:
            generic_item_class = _OPGenericItemRelaxedValidation
        else:
            generic_item_class = _OPGenericItem

        try:
            output = super()._item_get(document_identifier, vault=vault)
            item = generic_item_class(output)
        except OPItemGetException as e:
            raise OPDocumentDeleteException.from_opexception(e)
        # we want to return the explicit ID even if we were
        # given an document name or other identifier
        # that way the caller knows exactly what got deleted
        # can match it up with what they expected to be deleted, if desired
        document_id = item.unique_id

        # 'op document delete' doesn't have any stdout, so we're not
        # capturing any here
        self._document_delete(document_id, vault=vault, archive=archive)

        return document_id

    def item_list(self, categories=[], include_archive=False, tags=[], title_glob=None, vault=None, generic_okay=True) -> OPItemList:
        """
        Return a list of items in an account.

        Parameters
        ----------
        categories: List[str], optional
            A list of category names to restrict list to
        include_archive: bool, optional
            Include items in the Archive in the list
        tags: List[str], optional
            A list of tags to restrict list to
        title_glob: bool, optional
            a shell-style glob pattern to match against item titles. If provided,
            resulting list will include only matching items
            by default None
        vault: str, optional
            The name or ID of a vault to override the object's default vault
        generic_okay: bool, optional
            Instantiate unknown item types as _OPGenericItem rather than raise OPUnknownItemException

        Raises
        ------
        OPItemListException
            If the user list operation for any reason during command execution
        OPUnknownItemTypeException
            If thelist returned by 1Password contains one or more item descriptors
            that aren't a known type and generic_okay is False
        OPNotFoundException
            If the 1Password command can't be found

        Returns
        -------
        user: OPUserDescriptorList
            An object representing a list of user descriptors

        Service Account Support
        -----------------------
        Supported
        """
        item_list_json = self._item_list(
            categories, include_archive, tags, vault)
        item_list = OPItemList(item_list_json, generic_okay=generic_okay)

        if title_glob:
            _list = []
            for obj in item_list:
                if fnmatch.fnmatch(obj.title, title_glob):
                    _list.append(obj)
            item_list = OPItemList(_list)
        return item_list

    # TODO: Item creation is hard to test in an automated way since it results in changed
    #   state. There are operations during item creation that expect state to change from
    #   before to after item creation
    #   There is ongoing work in mock-op and mock-cli-framework to simulate changed state
    #   when this is complete these functions need to be tested
    #
    #   For now, ignore testing coverage
    def item_create(self,
                    new_item: OPNewItemMixin,
                    password_recipe: OPPasswordRecipe = None,
                    vault: Optional[str] = None):  # pragma: no coverage
        """
        Create a new item in the authenticated 1Password account

        Parameters
        ----------
        new_item: (OPNewItemMixin, OPAbstractItem)
            An object inheriting from OPnewItemMixin and OPAbstractItem representing the populated template
            of the new item to create
        password_recipe: OPPasswordRecipe
            Where appropriate, the password recipe to pass to '--password=' when creating the item
        vault: str
            Name of the vault to assign the item to
        Raises
        ------
        OPInvalidItemException
            - If new_item does not inherit from OPNewItemMixin
            - if password_recipe is provided and new_item does not support passwords
                (currently only Login and Password item types support passwords)
        OPItemCreateException
            If item creation fails for any reason during command execution

        Returns
        -------
        op_item: OPAbstractItem
            The newly created item object

        Service Account Support
        -----------------------
        Supported
          required keyword arguments: vault
        """
        if not isinstance(new_item, OPNewItemMixin):
            raise OPInvalidItemException(
                "Attempting to create item using object not from a template")

        # Most items don't support passwords. Only login & password items do
        # 'op' will fail if we provide a password recipe when creating an
        # an item that doesn't support passwords
        if password_recipe and not new_item.supports_passwords():
            raise OPInvalidItemException(
                "Password recpipe provided for an item that doesn't support passwords")
        result_str = super()._item_create(
            new_item, password_recipe=password_recipe, vault=vault)
        op_item = OPItemFactory.op_item(result_str)
        return op_item

    def login_item_create(self,
                          title: str,
                          username: str,
                          password: Union[str, OPPasswordRecipe] = None,
                          url: Optional[str] = None,
                          url_label: str = "Website",
                          tags: List[str] = [],
                          vault: str = None):  # pragma: no coverage
        """
        Create a new login item in the authenticated 1Password account

        Parameters
        ----------
        Title : str
            User viewable name of the login item to create
        username : str
            username string for the new login item
        password : Union[str, OPPasswordRecipe], optional
            May be one of:
                - the literal password string to set for this login item
                - a OPPasswordRecipe object that will be provided to '--generate-password='
            If a password string is provided that password will be set for this login item
            If an OPPasswordRecipe object is provided, it will ensure a well-formed password recipe string is provided to '--generate-password='
        url: str, optional
            If provided, set to the primary URL of the login item
        url_label: str, optional
            If provided and a URL is provided, this bcomes the primary URL's label
        tags: List[str], optional
            A list of tags to apply to the item when creating
        vault: str, optionsl
            The vault in which to create the new item

        Raises
        ------
        OPInvalidItemException
            - If new_item does not inherit from OPNewItemMixin
            - if password_recipe is provided and new_item does not support passwords
                (currently only Login and Password item types support passwords)
        OPItemCreateException
            If item creation fails for any reason during command execution

        Returns
        -------
        login_item: OPLoginItem
            The newly created login item object

        Service Account Support
        -----------------------
        Supported
          required keyword arguments: vault
        """
        password_recipe = None

        # if password is actually a password recipe,
        # set passsword_recipe and set password to None
        # that way we don't pass it into OPLoginItemTemplate
        # and instead pass it to _item_create() so it gets used on the command line
        if isinstance(password, OPPasswordRecipe):
            password_recipe = password
            password = None

        url_obj = None
        if url:
            url_obj = OPLoginItemNewPrimaryURL(url, url_label)

        new_item = OPLoginItemTemplate(
            title, username, password=password, url=url_obj, tags=tags)

        login_item = self.item_create(
            new_item, password_recipe=password_recipe, vault=vault)
        return login_item

    def item_delete(self, item_identifier: str, vault: Optional[str] = None, archive: bool = False, relaxed_validation=False) -> str:
        """
        Delete an item based on title or unique identifier

        Parameters
        ----------
        item_identifier: str
            Any item identifier accepted by 'op item delete'
            Generally this includes item name/title, item ID, or share link
            See 'op item get --help' for more information
        vault: str, optional
            The name or ID of a vault to override the default vault
        archive: bool
            Whether to archive or permanently delete the item

        Note:
            If a non-unique item identifier is provided (e.g., item name/title), and there
            is more than one item that matches, OPItemDeleteException will be raised. Check the
            error message in OPItemDeleteException.err_output for details

        Raises
        ------
        OPItemDeleteException
            - If the item to be deleted is not found
            - If there is more than one item matching 'item_identifier'
            - If the delete operation fails for any other reason

        Returns
        -------
        item_id: str
            Unique identifier of the item deleted

        Service Account Support
        -----------------------
        Supported
          required keyword arguments: vault
        """

        # to satisfy mypy
        generic_item_class: Type[_OPGenericItem]
        if relaxed_validation:
            generic_item_class = _OPGenericItemRelaxedValidation
        else:
            generic_item_class = _OPGenericItem

        try:
            output = super()._item_get(item_identifier, vault=vault)
            item = generic_item_class(output)
        except OPItemGetException as e:
            raise OPItemDeleteException.from_opexception(e)
        # we want to return the explicit ID even if we were
        # given an item title or other identifier
        # that way the caller knows exactly what got deleted
        # can match it up with what they expected to be deleted, if desired
        item_id = item.unique_id

        # 'op item delete' doesn't have any stdout, so we're not
        # capturing any here
        self._item_delete(item_id, vault=vault, archive=archive)

        return item_id

    def item_delete_multiple(self,
                             vault,
                             categories=[],
                             include_archive=False,
                             tags=[],
                             archive=False,
                             title_glob=None,
                             batch_size=25):
        """
        Delete multiple items at once from a specific vault. This may take place across
        one or more 'op item delete' passes. A maximum "batch size" number of items to
        delete in each pass may optionally be specified (defaulting to 25)

        Parameters
        ----------
        vault: str
            The name or ID of a vault to delete from. This parameter is mandatory to mitigate
            the risk of deleting many items from the wrong vault
        include_archive: bool, optional
            Whether to include archived items for deleting
            by default False
        archive: bool
            Whether to archive or permanently delete the items
            by default False
        tags: List[str], optional
            A list of tags to restrict batch deletion to
        title_glob: bool, optional
            a shell-style glob pattern to match against item titles for deleting
            by default None
        batch_size: int, optional
            Maximum number of items to delete in each pass
            by default 25
            NOTE: The default batch size is subject to change without notice

        Note:
            If a non-unique item identifier is provided (e.g., item name/title), and there
            is more than one item that matches, OPItemDeleteException will be raised. Check the
            error message in OPItemDeleteException.err_output for details

        Raises
        ------
        OPItemDeleteMultipleException
            - If the 'item list' operation fails
            - If any of the 'item delete' operations fail

        Returns
        -------
        item_id: str
            Unique identifier of the item deleted

        Service Account Support
        -----------------------
        Supported
          required keyword arguments: vault
        """
        # track deleted items as we delete them so we can return
        # that list to the caller
        deleted_items = OPItemList([])

        try:
            item_list = self.item_list(categories=categories,
                                       include_archive=include_archive,
                                       tags=tags,
                                       title_glob=title_glob,
                                       vault=vault)
        except OPItemListException as e:
            raise OPItemDeleteMultipleException.from_opexception(
                e, deleted_items)

        batches: List[OPItemList] = []
        start = 0
        end = len(item_list)

        for i in range(start, end, batch_size):
            # split item list into chunks >= batch_size each
            x = i
            chunk = item_list[x:x+batch_size]
            batches.append(OPItemList(chunk))

        for batch in batches:
            batch_json = batch.serialize()
            try:
                self._item_delete_multiple(batch_json, vault, archive=archive)
            except OPCmdFailedException as ope:  # pragma: no coverage
                raise OPItemDeleteMultipleException.from_opexception(
                    ope, deleted_items)
            deleted_items.extend(batch)

        return deleted_items

    def signed_in_accounts(self, decode="utf-8") -> OPAccountList:
        """
        Retrieve a users and accounts set up on this device

        Returns
        -------
        OPAccountList
            List of accounts (subclass of List)

        Service Account Support
        -----------------------
        Supported
        """
        account_list_json = super()._signed_in_accounts(self.op_path, decode=decode)
        account_list = OPAccountList(account_list_json)
        return account_list

    def signout(self, forget=False):  # pragma: no coverage
        """
        Sign out of the account used to create this OP instance
        This is equivalent to the command 'op signout'
        Parameters
        ----------
        forget: bool, optional
            Optionally remove details for this 1Password account from this device.
            This is equivalent to the command 'op signout --forget'

        Raises
        ------
        OPSignoutException
            If the signout operation fails for any reason
        OPNotFoundException
            If the 1Password command can't be found

        Returns
        -------
        None

        Service Account Support
        -----------------------
        Supported
            Note: Has no effect on the authentication status of the service account currently in use
        """
        account = self._account_identifier
        token = self.token
        if not token and not self._uses_bio:
            return

        try:
            super()._signout(account, token, forget=forget)
        except OPCmdFailedException as ocfe:
            raise OPSignoutException.from_opexception(ocfe) from ocfe

        # drop any reference to op session token identifier from this
        # instance and from environment variables
        self._sanitize()

    @classmethod
    def account_forget(cls, account: str, op_path=None):  # pragma: no coverage
        """
        Remove a 1Password account from this device
        This is equivalent to the command 'op account forget <account_id>'

        Note: this is a class method, so there is no need to have an OP instance or to have
        an active, signed-in session

        Note 2: this only removes accounts that have been added using 'op account add'.
        It has no effect on accounts accessed through 1Password app integration

        Parameters
        ----------
        account : str
            The account shorthand to forget
        op_path: str, optional
            Path to an 'op' executable to use for this action

        Raises
        ------
        OPForgetException
            If the lookup fails for any reason
        OPNotFoundException
            If the 1Password command can't be found

        Returns
        -------
        None

        Service Account Support
        -----------------------
        Not supported
        """

        try:
            cls._account_forget(account, op_path=op_path)
        except OPCmdFailedException as ocfe:
            raise OPForgetException.from_opexception(ocfe) from ocfe

    def _sanitize(self):  # pragma: no coverage
        self._token = None
        if self._sess_var:
            try:
                env.pop(self._sess_var)
            except KeyError:
                pass



from ._item_type_registry import OPItemFactory, OPUnknownItemTypeException
from .fields_sections._new_fields import OPNewConcealedField, OPNewStringField
from .item_types._item_base import OPFieldNotFoundException



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from pyonepassword import OP
    from .fixtures.expected_vault_data import (
        ExpectedVaultData,
        ExpectedVault
    )

from pyonepassword.api.exceptions import (
    OPInvalidVaultException,
    OPVaultGetException
)
from pyonepassword.api.object_types import OPVault

# ensure HOME env variable is set, and there's a valid op config present
pytestmark = pytest.mark.usefixtures("valid_op_cli_config_homedir")


def test_vault_get_01(signed_in_op: OP, expected_vault_data: ExpectedVaultData):
    # get vault "Test Data"
    expected: ExpectedVault
    result: OPVault

    vault_name = "Test Data"
    expected = expected_vault_data.data_for_vault(vault_name)
    result = signed_in_op.vault_get(vault_name)

    assert isinstance(result, OPVault)
    assert result.unique_id == expected.unique_id


def test_vault_get_invalid_01(signed_in_op: OP, expected_vault_data):
    vault_name = "Invalid Vault"
    expected = expected_vault_data.data_for_vault(vault_name)
    try:
        _ = signed_in_op.vault_get(vault_name)
        assert False, "We should have caught an exception"
    except OPVaultGetException as e:
        print(e)
        assert e.returncode == expected.returncode


def test_vault_get_malformed_json_01(invalid_data):
    malformed_json = invalid_data.data_for_name("malformed-vault-json")
    with pytest.raises(OPInvalidVaultException):
        OPVault(malformed_json)



from ..op_items.item_validation_policy import (
    disable_relaxed_validation,
    enable_relaxed_validation,
    get_relaxed_validation,
    get_relaxed_validation_for_class,
    set_relaxed_validation_for_class,
    set_strict_validation_for_class
)

__all__ = ["enable_relaxed_validation",
           "disable_relaxed_validation",
           "get_relaxed_validation",
           "get_relaxed_validation_for_class",
           "set_relaxed_validation_for_class",
           "set_strict_validation_for_class"]



from . import __summary__, __title__, __version__


class PyOPAbout:
    def __init__(self) -> None:
        self.version = __version__
        self.summary = __summary__
        self.title = __title__

    def __str__(self):
        return f"{self.title.upper()}: {self.summary}. Version {self.version}"


class PyOPAboutMixin:
    _about = PyOPAbout()

    @classmethod
    def about(cls) -> str:
        """
        Class method to generate an "about" summary string for pyonepassword, which includes:
            - project title
            - project summary
            - project version

        Returns
        -------
        str
            The about string
        """
        return str(cls._about)

    @classmethod
    def version(cls) -> str:
        """
        Class method to return the version string for pyonepassword, e.g., 3.4.0

        Returns
        -------
        str
            The version string
        """
        return cls._about.version



import logging
import subprocess
from os import environ

from .py_op_exceptions import (
    OPCLIPanicException,
    OPCmdFailedException,
    OPNotFoundException,
    OPRevokedSvcAcctTokenException
)

# Mainly for use in automated testing
LOG_OP_ERR_ENV_NAME = "LOG_OP_ERR"

"""
Module to hold stuff that interacts directly with 'op' or its config

TODO: Move other code that closely touches 'op' here
"""


class _OPCLIExecute:
    # we need to detect if a command failure was actually a mock-op failure
    MOCK_OP_ERR_EXIT = 255
    MOCK_OP_RESP_ERR_MSG = "Error looking up response"
    GO_RUNTIME_PANIC_MSG = "panic: runtime error:"
    SVC_ACCT_REVOKED_MSG = "The Service Account used in this integration has been deleted"
    logger = logging.getLogger("_OPCLIExecute")
    logger.setLevel(logging.INFO)

    def __new__(cls, *args, logger=None, **kwargs):
        if logger:
            cls.set_logger(logger)
        return super().__new__(cls)
    """
    Class for logging into and querying a 1Password account via the 'op' cli command.
    """

    @classmethod
    def _should_log_op_errors(cls) -> bool:
        should_log = False
        if environ.get(LOG_OP_ERR_ENV_NAME) == "1":
            should_log = True
        return should_log

    @classmethod
    def _run_raw(cls, argv, input_string=None, capture_stdout=False, ignore_error=False, env=environ):
        stdout = subprocess.PIPE if capture_stdout else None
        if input_string:
            if isinstance(input_string, str):
                input_string = input_string.encode("utf-8")

        _ran = subprocess.run(
            argv, input=input_string, stderr=subprocess.PIPE, stdout=stdout, env=env)

        stdout = _ran.stdout
        stderr = _ran.stderr
        returncode = _ran.returncode

        if not ignore_error:
            try:
                _ran.check_returncode()
            except subprocess.CalledProcessError as err:
                stderr_output = stderr.decode("utf-8").rstrip()
                if cls._should_log_op_errors():
                    cls.logger.error(f"'op' command error: {stderr_output}")
                # HACK:
                # mock-op returns -1 (i.e., 255) if it can't find a response
                # but op (currently) only ever returns 1 on error
                #
                # we need to check if this was a mock-op failure so that during
                # testing we can distinguish between a simulated 'op' command failure
                # and mock-op failing because we haven't provided an appropriate response
                # definition
                if (returncode >= cls.MOCK_OP_ERR_EXIT and
                        cls.MOCK_OP_RESP_ERR_MSG in stderr_output):  # pragma: no coverage
                    raise err
                elif cls.GO_RUNTIME_PANIC_MSG in stderr_output:
                    # If we made 'op' crash, raise a special exception
                    raise OPCLIPanicException(stderr_output, returncode, argv)
                elif cls.SVC_ACCT_REVOKED_MSG in stderr_output:
                    # do this unconditionally without checking if we're authed as
                    # a service account
                    # in case caller is accidentally running with OP_SERVICE_ACCOUNT_TOKEN
                    raise OPRevokedSvcAcctTokenException(
                        stderr_output, returncode)

                raise OPCmdFailedException(stderr_output, returncode) from err

        return (stdout, stderr, returncode)

    @classmethod
    def _run(cls, argv, capture_stdout=False, input_string=None, decode=None, env=environ):
        cls.logger.debug(f"Running: {argv.cmd_str()}")
        output = None
        try:
            output, _, _ = cls._run_raw(
                argv, input_string=input_string, capture_stdout=capture_stdout, env=env)
            if decode and output is not None:
                output = output.decode(decode)
        except FileNotFoundError as err:
            cls.logger.error(
                "1Password 'op' command not found at: {}".format(argv[0]))
            cls.logger.error(
                "See https://developer.1password.com/docs/cli for more information")
            raise OPNotFoundException(argv[0], err.errno) from err

        return output

    @classmethod
    def set_logger(cls, logger: logging.Logger):
        cls.logger = logger

    @classmethod
    def set_log_level(cls, log_level: int):  # pragma: no coverage
        cls.logger.setLevel(log_level)



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from pyonepassword.api.object_types import OPAccount

if TYPE_CHECKING:
    from pyonepassword import OP

    from .fixtures.expected_account_data import ExpectedAccountData


pytestmark = pytest.mark.usefixtures("valid_op_cli_config_homedir")


def test_account_list_01(signed_in_op: OP, expected_account_data: ExpectedAccountData):
    account_id = "user1@yahoo.com"
    account_list = signed_in_op.signed_in_accounts()
    account: OPAccount = account_list.account_for_identifier(account_id)
    assert isinstance(account, OPAccount)


def test_account_list_02(signed_in_op: OP, expected_account_data: ExpectedAccountData):
    account_id = "user1@yahoo.com"
    expected = expected_account_data.data_for_account(account_id)

    account_list = signed_in_op.signed_in_accounts()
    account: OPAccount = account_list.account_for_identifier(account_id)

    assert account.user_uuid == expected.user_uuid


def test_account_list_03(signed_in_op: OP, expected_account_data: ExpectedAccountData):
    account_id = "user2@yahoo.com"
    expected = expected_account_data.data_for_account(account_id)

    account_list = signed_in_op.signed_in_accounts()
    account: OPAccount = account_list.account_for_identifier(account_id)

    assert account.email == expected.email


def test_account_list_04(signed_in_op: OP, expected_account_data: ExpectedAccountData):
    account_id = "user2@yahoo.com"
    expected = expected_account_data.data_for_account(account_id)

    account_list = signed_in_op.signed_in_accounts()
    account: OPAccount = account_list.account_for_identifier(account_id)

    assert account.account_uuid == expected.account_uuid


def test_account_list_05(signed_in_op: OP, expected_account_data: ExpectedAccountData):
    account_id = "user1@yahoo.com"
    expected = expected_account_data.data_for_account(account_id)

    account_list = signed_in_op.signed_in_accounts()
    account: OPAccount = account_list.account_for_identifier(account_id)

    assert account.url == expected.url



import json

from .. import data
from ..pkg_resources import pkgfiles


class OPTemplateDirectory:

    def __init__(self):
        with pkgfiles(data).joinpath(data.TEMPLATE_REGISTRY_JSON).open("r") as _file:
            self._registry = json.load(_file)

    def template_for_category(self, category: str):
        template_name = self._registry[category]
        template = None
        with pkgfiles(data).joinpath(template_name).open("r") as _file:
            template = json.load(_file)

        return template



from .._item_descriptor_registry import op_register_item_descriptor_type
from ._item_descriptor_base import OPAbstractItemDescriptor


@op_register_item_descriptor_type
class OPIdentityItemDescriptor(OPAbstractItemDescriptor):
    CATEGORY = "IDENTITY"

    def __init__(self, item_dict):
        super().__init__(item_dict)



from .._item_descriptor_registry import op_register_item_descriptor_type
from .._item_type_registry import op_register_item_type
from ._item_base import OPAbstractItem
from ._item_descriptor_base import OPAbstractItemDescriptor

# {
#     "id": "2dvgl7kk5yjrq3gxwqimp5awve",
#     "title": "Example Password",
#     "version": 1,
#     "vault": {
#         "id": "gshlsjsajnawtnjynzgwmiebge",
#         "name": "Test Data"
#     },
#     "category": "PASSWORD",
#     "last_edited_by": "5GHHPJK5HZC5BAT7WDUXW57G44",
#     "created_at": "2021-11-23T05:46:01Z",
#     "updated_at": "2021-11-23T05:46:01Z",
#     "additional_information": "2021-11-22 09:46 PM",
#     "fields": [
#         {
#             "id": "password",
#             "type": "CONCEALED",
#             "purpose": "PASSWORD",
#             "label": "password",
#             "value": "iXdx8KAEmUkPqCvHHYjngHzRr7",
#             "reference": "op://Test Data/Example Password/password",
#             "password_details": {
#                 "strength": "FANTASTIC"
#             }
#         },
#         {
#             "id": "notesPlain",
#             "type": "STRING",
#             "purpose": "NOTES",
#             "label": "notesPlain",
#             "reference": "op://Test Data/Example Password/notesPlain"
#         }
#     ]
# }


@op_register_item_descriptor_type
class OPPasswordItemDescriptor(OPAbstractItemDescriptor):
    CATEGORY = "PASSWORD"

    def __init__(self, item_dict):
        super().__init__(item_dict)


@op_register_item_type
class OPPasswordItem(OPAbstractItem):
    CATEGORY = "PASSWORD"

    def __init__(self, item_dict):
        super().__init__(item_dict)

    @property
    def password(self):
        password = self.field_value_by_id("password")
        return password


@op_register_item_type
class OPPasswordItemRelaxedValidation(OPPasswordItem):
    _relaxed_validation = True



import pyonepassword.api.authentication
import pyonepassword.api.constants
import pyonepassword.api.decorators
import pyonepassword.api.descriptor_types
import pyonepassword.api.exceptions
import pyonepassword.api.object_types
import pyonepassword.api.validation

"""
Test that all symbols exported from pyonepassword.api are properly re-exported
"""


def test_authentication_exports():
    """
    Verify all symbols in pyonepassword.api.authentication are properly re-exported
    """
    authentication_all = pyonepassword.api.authentication.__all__
    for symbol in dir(pyonepassword.api.authentication):
        if symbol.startswith("__"):
            continue
        assert symbol in authentication_all


def test_constants_exports():
    """
    Verify all symbols in pyonepassword.api.constants are properly re-exported
    """
    constants_all = pyonepassword.api.constants.__all__
    for symbol in dir(pyonepassword.api.constants):
        if symbol.startswith("__"):
            continue
        assert symbol in constants_all


def test_decorators_exports():
    """
    Verify all symbols in pyonepassword.api.decorators are properly re-exported
    """
    decorators_all = pyonepassword.api.decorators.__all__
    for symbol in dir(pyonepassword.api.decorators):
        if symbol.startswith("__"):
            continue
        assert symbol in decorators_all


def test_descriptor_types_exports():
    """
    Verify all symbols in pyonepassword.api.descriptor_types are properly re-exported
    """
    descriptor_types_all = pyonepassword.api.descriptor_types.__all__
    for symbol in dir(pyonepassword.api.descriptor_types):
        if symbol.startswith("__"):
            continue
        assert symbol in descriptor_types_all


def test_exceptions_exports():
    """
    Verify all symbols in pyonepassword.api.exceptions are properly re-exported
    """
    exceptions_all = pyonepassword.api.exceptions.__all__
    for symbol in dir(pyonepassword.api.exceptions):
        if symbol.startswith("__"):
            continue
        assert symbol in exceptions_all


def test_object_types_exports():
    """
    Verify all symbols in pyonepassword.api.object_types are properly re-exported
    """
    object_types_all = pyonepassword.api.object_types.__all__
    for symbol in dir(pyonepassword.api.object_types):
        if symbol.startswith("__"):
            continue
        assert symbol in object_types_all


def test_object_validation_exports():
    """
    Verify all synmbols in pyonepassword.api.validation are properly re-exported
    """
    assert hasattr(pyonepassword.api.validation, "__all__")
    validation_all = pyonepassword.api.validation.__all__
    for symbol in dir(pyonepassword.api.validation):
        if symbol.startswith("__"):
            continue
        assert symbol in validation_all



from typing import Set

"""
Module for classes & functions related to item valiation policy.

In some cases `op` may return item JSON that is not strictly conforming.
Examples include fields with empty-string IDs, duplicate sections, and
other anomalies.

In this case we have no control of the anomalous data, so we must relax validation
in order to instantiate useful item objects without raising exceptions

This module provides mechanisms to enable/disable relaxed validation globaly, or on a
per-class basis
"""


class _OPItemValidationPolicy:
    """
    Class representing the active policy whether a 1Password item object should
    enforce strict validation policy when processing a dictionary from JSON
    """
    _relaxed_item_classes: Set[type] = set()
    _relaxed_validation: bool = False

    @classmethod
    def _enable_relaxed_validation(cls) -> None:
        """
        Enable relaxed validation policy globally
        """
        cls._relaxed_validation = True

    @classmethod
    def _disable_relaxed_validation(cls) -> None:
        """
        Disable relaxe validation policy globally

        Per-class relaxed validation will still apply
        """
        cls._relaxed_validation = False

    @classmethod
    def _get_relaxed_validation(cls, item_class: type = None) -> bool:
        """
        Get the validation policy taking into a account global policy and optionally 'item_class':
            True if either is true

        Note: if item_class is not provided, then only global validation policy is consulted

        Parameters
        ----------
        item_class : type, optional
            Any OPAbstractItem class, by default None

        Returns
        -------
        bool
            Whether the relaxed validation policy is enabled
        """
        global_relaxed = cls._relaxed_validation

        class_relaxed = cls._get_relaxed_validation_for_class(item_class)

        # Union of global and class policy
        relaxed = global_relaxed or class_relaxed

        return relaxed

    @classmethod
    def _get_relaxed_validation_for_class(cls, item_class: type) -> bool:
        """
        Get the validation policy only for 'item_class'

        Parameters
        ----------
        item_class : type
            Any OPAbstractItem class

        Returns
        -------
        bool
            Whether relaxed validation is set for 'item_class'
        """
        relaxed = False
        # is obj_or_class, one of the classes (or a subclass of)
        # for which validation is relaxed?
        if issubclass(item_class, tuple(cls._relaxed_item_classes)):
            relaxed = True
        return relaxed

    @classmethod
    def _set_relaxed_validation_for_class(cls, item_class) -> None:
        """
        Enable relaxed validation policy for 'item_class'

        Parameters
        ----------
        item_class : type
            Any OPAbstractItem class
        """
        cls._relaxed_item_classes.add(item_class)

    @classmethod
    def _set_strict_validation_for_class(cls, item_class) -> None:
        """
        Disable relaxed validation policy for 'item_class'

        Parameters
        ----------
        item_class : type
            Any OPAbstractItem class
        """
        if item_class in cls._relaxed_item_classes:
            cls._relaxed_item_classes.remove(item_class)


def enable_relaxed_validation() -> None:
    """
    Convenience method to enable relaxed validation policy globally
    """
    _OPItemValidationPolicy._enable_relaxed_validation()


def disable_relaxed_validation() -> None:
    """
    Convenience method to disable relaxed validation policy globally.

    Per-class relaxed validation policy still applies
    """
    _OPItemValidationPolicy._disable_relaxed_validation()


def get_relaxed_validation(item_class=None) -> bool:
    """
    Get relaxed validation policy.

    If optional 'item_class' is provided, the returned value represents the union of global
    item validation policy, and the class-specific item validation policy.

    This is to say True is returned if either is true

    Parameters
    ----------
    item_class : type, optional
        Any OPAbstractItem class, by default None

    Returns
    -------
    bool
        The union of the global item validation policy and the class's item validation policy
    """
    return _OPItemValidationPolicy._get_relaxed_validation(item_class=item_class)


def get_relaxed_validation_for_class(item_class) -> bool:
    """
    Get item validation policy for a specific op item class

    Parameters
    ----------
    item_class : type
        Any OPAbstractItem class

    Returns
    -------
    bool
        True of relaxed validation policy is set for this class
    """
    return _OPItemValidationPolicy._get_relaxed_validation_for_class(item_class)


def set_relaxed_validation_for_class(item_class) -> None:
    """
    Set item validation policy for the specified op item class

    Parameters
    ----------
    item_class : type
        Any OPAbstractItem class
    """
    _OPItemValidationPolicy._set_relaxed_validation_for_class(item_class)


def set_strict_validation_for_class(item_class) -> None:
    """
    Remove the specified op item class from the relaxed validation list

    Note: relaxed validation may still occur if it has been set globally

    Parameters
    ----------
    item_class : type
        Any OPAbstractItem class
    """
    _OPItemValidationPolicy._set_strict_validation_for_class(item_class)



import enum
import json
from typing import Any, Dict, List, NamedTuple, Optional, Set

from pysingleton import PySingleton  # type: ignore

from . import data
from .pkg_resources import data_location_as_path

OptionalStrList = Optional[List[str]]


class OPSvcAcctCommandNotSupportedException(Exception):
    pass


class OPSvcAcctSupportedEnum(enum.IntEnum):
    NOT_VALIDATED = -1
    SUPPORTED = 0  # the command is supported by service accounts as-is
    INCOMPAT_OPTIONS = 1  # the command is supported by service accounts,
    # but the required vault argument is missing
    NOT_SUPPORTED = 2  # the command is not supported by service accounts


SVC_ACCT_SUPPORTED = OPSvcAcctSupportedEnum.SUPPORTED
SVC_ACCT_INCOMPAT_OPTIONS = OPSvcAcctSupportedEnum.INCOMPAT_OPTIONS
SVC_ACCT_CMD_NOT_SUPPORTED = OPSvcAcctSupportedEnum.NOT_SUPPORTED
_SVC_ACCT_CMD_NOT_VALIDATED = OPSvcAcctSupportedEnum.NOT_VALIDATED


class OPSvcAcctSupportCode(NamedTuple):
    code: OPSvcAcctSupportedEnum
    msg: Optional[str]


class _CmdSpec(dict):

    def __init__(self, cmd_dict):
        super().__init__(cmd_dict)

    @property
    def command_name(self) -> str:
        return self["meta"]["command_name"]

    def _main_command_dict(self) -> Dict:
        # Return command dict for the top-level command
        # if there is one
        # else empty dict
        cmd_dict = self.get("cmd_dict", {})
        return cmd_dict

    def command_lookup(self, subcommand_list: OptionalStrList = None):
        # subcommand_list, if provided, is a subcommand chain of one or more items
        # e.g, the "op item" command has the following subcommands (not exhaustive):
        # - get (e.g., op item get <item>)
        # - template list (e.g., op item template list <item_type>)
        subcmd_dict = self.get("subcommands", {})
        found: Dict[Any, Any] = {}

        if subcommand_list:
            # if we were given a subcommand list with one or more items
            # then try to find the command dict for the subcommand chain
            for subcmd in subcommand_list:
                subcmd_dict = subcmd_dict.get(subcmd, {})
                if "has_arg" in subcmd_dict:
                    # check if current subcommand is the most deeply nested,
                    # if so, the next entry in our list may be a positional
                    # argument argument and not another subcommand,
                    # so stop processing here
                    found = subcmd_dict
                    break
        else:
            # there was no subcommand list (e.g., in the case of op whoami)
            # so get the top-level command dict
            found = self._main_command_dict()

        return found


# Make the registry a singleton to avoid
# loading all the JSON files from disk each time
# it gets instantiated
class OPSvcAcctSupportRegistry(metaclass=PySingleton):
    # This allows the instance to be freed & reinitialized
    # once the last reference released
    _PYSINGLETON_WEAKREF = True
    """
    A registry of supported commands, subcommands, and required & prohibited options
    in the context of service accounts
    """

    def __init__(self):
        self._supported_commands = {}
        # data_location_as_path() satisfies mypy
        # by returning a Path instead of a Traversable
        data_path = data_location_as_path(data, data.SVC_ACCOUNT_COMMANDS)

        for json_file in data_path.glob("*.json"):
            cmd_dict = json.load(open(json_file, "r"))
            self._process(cmd_dict)

    def _process(self, cmd_dict: Dict):
        cmd_spec = _CmdSpec(cmd_dict)
        cmd = cmd_spec.command_name
        self._supported_commands[cmd] = cmd_spec

    def command_supported(self, _argv: List[str]) -> OPSvcAcctSupportCode:
        # This function is much more complex than I'd like, but most of the complexity is
        # around building necessary context for a meaningful exception message

        _support_code = _SVC_ACCT_CMD_NOT_VALIDATED
        _support_msg: str = None

        # for testing of all required options are provided
        # and for generating error message
        reqd_opt_diff: Set[str] = set()
        required_opt_satified = False
        # for generating error message if any prohibted options are found
        prohib_opt_diff: Set[str] = set()
        prohibited_opt_satisfied = False
        # copy argv so we don't modify the original
        argv_list: List[str] = list(_argv)
        # [op_exe, [global options, ...], [subcommands, ...], [--sub-cmd-options, ...]]
        cmd_spec: _CmdSpec

        # pop argv[0], the op exe path
        argv_list.pop(0)

        # the following global options each take an argument
        # e.g., "--format json"
        global_options_with_args = ["--format",
                                    "--encoding", "--session", "--config"]

        # pop global options
        while argv_list and argv_list[0].startswith("--"):
            arg = argv_list.pop(0)
            # does the option have an argument that also
            # needs to be popped?
            # e.g., if we popped "--format", we also
            # need to pop its argument, "json"
            if arg in global_options_with_args:
                argv_list.pop(0)

        # get primary command
        command = None
        if argv_list:
            command = argv_list.pop(0)

        # build subcommands:
        subcommands = []
        while argv_list and not argv_list[0].startswith("--"):
            subcommands.append(argv_list.pop(0))

        # build subcommand option list
        subcmd_options = []
        while argv_list:
            # pop off all subcommand options, ignoring option-arguments
            # e.g, save "--vault", but skip argument "Test Data"
            option = argv_list.pop(0)
            if option.startswith("--"):
                subcmd_options.append(option)

        if command is None:
            # command-less options such as --version are always supported
            _support_code = SVC_ACCT_SUPPORTED
            _support_msg = None
            cmd_spec = None
            required_opt_satified = True
            prohibited_opt_satisfied = True
        else:
            cmd_spec = self._supported_commands.get(command)

        cmd_dict = {}
        if cmd_spec:
            cmd_dict = cmd_spec.command_lookup(subcommand_list=subcommands)

        # if we either failed to find a command spec or
        # we found a command spec but not a subcommand dictionary
        # then command is not supported
        _cmd_not_found = command is not None and cmd_spec is None
        if (cmd_spec and not cmd_dict) or _cmd_not_found:
            _support_msg = f"Command or subcommand not supported: [{command} {' '.join(subcommands)}]"
            _support_code = SVC_ACCT_CMD_NOT_SUPPORTED

        if cmd_dict and _support_code == _SVC_ACCT_CMD_NOT_VALIDATED:
            required_options = set(cmd_dict["required_options"])

            # how many required options remain after we mask out
            # the provided options? Hopefully none
            # we use the diff for two purposes:
            # 1. to test if any required options were omitted
            # 2. to generate a meaninful error message if needed
            reqd_opt_diff = required_options - set(subcmd_options)

            if not reqd_opt_diff:
                # all required options were in use, so we're good here
                required_opt_satified = True
            else:
                _support_code = SVC_ACCT_INCOMPAT_OPTIONS

            prohibited_options = set(cmd_dict["prohibited_options"])

            # how many probited options remain after we mask out
            # the provided options? Hopefully all of them
            remaining = prohibited_options - set(subcmd_options)

            if remaining == prohibited_options:
                # didn't use any prohbited options, so we're good here
                prohibited_opt_satisfied = True
            else:
                # which prohibted options were actually used?
                # we need this to generate the error message
                prohib_opt_diff = prohibited_options - remaining
                _support_code = SVC_ACCT_INCOMPAT_OPTIONS

            if required_opt_satified and prohibited_opt_satisfied:
                # if both
                _support_code = SVC_ACCT_SUPPORTED

        if _support_code == SVC_ACCT_INCOMPAT_OPTIONS:
            _support_msg = ""
            if not required_opt_satified:
                _support_msg += f"Required options not provided: [{','.join(list(reqd_opt_diff))}]"

            if not prohibited_opt_satisfied:
                _support_msg += f" Prohibited options found: [{','.join(list(prohib_opt_diff))}]"

            _support_msg = _support_msg.lstrip()

        if _support_code == _SVC_ACCT_CMD_NOT_VALIDATED:  # pragma: no cover
            # something's gone wrong if we reach this point
            raise Exception("Failed to validate service account compatibility")

        supported = OPSvcAcctSupportCode(_support_code, _support_msg)

        return supported



import json


def safe_unjson(json_or_obj):
    """
    Transparently un-json things if they are strings, and no-op if not
    """
    if isinstance(json_or_obj, str):
        obj = json.loads(json_or_obj)
    else:
        obj = json_or_obj
    return obj



# Most of the object types and descriptor types need to
# register themselves with a factory class
# so we need to import them here to ensure that happens
from .api.decorators import *  # noqa: F401, F403
from .api.descriptor_types import *  # noqa: F401, F403
from .api.object_types import *  # noqa: F401, F403

# Declare token object that can be imported into __init__.py without polluting namespace
# to ensure the classes mentioned above get imported & initialized
_API_INITIALIZED = True
